"""
Class for loading data into Pytorch float tensor

From: https://gitlab.com/acasamitjana/latentmodels_ad

"""
import numpy as np
import torch
from torch.utils.data import Dataset
from os.path import join

class MVDataset(Dataset):
    """PyTorch Dataset for storing and accessing multi-view data.

    Args:
        data (list): Input data. list of torch.Tensors.
        labels (np.array): Dataset labels. 
        return_index (bool): Whether to return batch index labels.
        transform (torchvision.transforms): Torchvision transformation to apply to the data. Default is None.
    """
    def __init__(self, data, n_views, labels=None, return_index=False, transform=None):

        self.data = data
        self.labels = labels
        self.n_views = n_views
        self.return_index = return_index
        self.transform = transform

        self.N = len(self.data[0])
        self.data = [
            torch.from_numpy(d).float() if isinstance(d, np.ndarray) else d
            for d in self.data
        ]
        self.shape = [
            np.shape(d) for d in self.data
        ]

        if labels is not None:
            self.labels = torch.from_numpy(self.labels)
     

    def __getitem__(self, index):

        x = [d[index] for d in self.data]
        if self.transform:
            x = self.transform(x)

        if self.return_index:
            if self.labels is not None:
                return x, self.labels[index], index
            return x, index

        if self.labels is not None:
            return x, self.labels[index]
        return x

    def __len__(self):
        return self.N

class IndexMVDataset(Dataset):

    def __init__(self, data, n_views, labels=None, filename="", data_dir=''):
        self.N = len(data[0]) 
        self.n_views = n_views

        self.filename = filename
        self.data_dir = data_dir
        
        self.data = data[0]
        self.labels = labels

    def __getitem__(self, index):
        """
        Args:
            index (int): Index
        Returns:
            list (): returns list of numpy arrays
        """
        x = []
        idx = self.data[index]
        for i in range(self.n_views):
            x_i = np.load(join(self.data_dir, self.filename.format(i, idx)))
            x_i = torch.from_numpy(x_i).float()
            x.append(x_i)
        return x

    def __len__(self):
        return self.N
    




import ast
import torch
import hydra

import torch.nn as nn
import torch.nn.functional as F

from torch.nn import Parameter

class Encoder(nn.Module):
    """MLP Encoder

    Args:
        input_dim (list): Dimensionality of the input data.
        z_dim (int): Number of latent dimensions.
        hidden_layer_dim (list): Number of nodes per hidden layer.
        non_linear (bool): Whether to include a ReLU() function between layers.
        bias (bool): Whether to include a bias term in hidden layers.
        enc_dist (multiviewae.base.distributions.Default): Encoder distribution.
    """
    def __init__(
        self,
        input_dim,
        z_dim,
        hidden_layer_dim,
        non_linear,
        bias,
        enc_dist
    ):
        super().__init__()

        self.input_size = input_dim
        self.z_dim = z_dim
        self.hidden_layer_dim = hidden_layer_dim
        self.bias = bias
        self.enc_dist = enc_dist
        self.non_linear = non_linear
        self.layer_sizes = [input_dim] + self.hidden_layer_dim + [z_dim]
        lin_layers = [
            nn.Linear(dim0, dim1, bias=self.bias)
            for dim0, dim1 in zip(
                self.layer_sizes[:-1], self.layer_sizes[1:]
            )
        ]

        self.encoder_layers = nn.Sequential(*lin_layers)

    def forward(self, x):
        h1 = x
        for it_layer, layer in enumerate(self.encoder_layers[0:-1]):
            h1 = layer(h1)
            if self.non_linear:
                h1 = F.relu(h1)
        h1 = self.encoder_layers[-1](h1)

        return h1

class VariationalEncoder(Encoder):
    """Variational MLP Encoder

    Args:
        input_dim (list): Dimensionality of the input data.
        z_dim (int): Number of latent dimensions.
        hidden_layer_dim (list): Number of nodes per hidden layer.
        non_linear (bool): Whether to include a ReLU() function between layers.
        bias (bool): Whether to include a bias term in hidden layers.
        sparse (bool): Whether to enforce sparsity of the encoding distribution.
        log_alpha (float): Log of the dropout parameter.
        enc_dist (multiviewae.base.distributions.Normal, multiviewae.base.distributions.MultivariateNormal): Encoder distribution.
        multiple_latents (bool, optional): Whether the model using a separate linear layers for shared and private latent spaces.
        u_dim (int, optional): Dimensionality of the shared latent space.
        w_dim (int, optional): Dimensionality of the private latent space.
    """
    def __init__(
        self,
        input_dim,
        z_dim,
        hidden_layer_dim,
        non_linear,
        bias,
        sparse,
        log_alpha,
        enc_dist, 
        multiple_latents=False,
        u_dim=None,
        w_dim=None
    ):
        super().__init__(input_dim=input_dim,
                        z_dim=z_dim,
                        hidden_layer_dim=hidden_layer_dim,
                        bias=bias,
                        non_linear=non_linear,
                        enc_dist=enc_dist)

        self.sparse = sparse
        self.non_linear = non_linear
        self.log_alpha = log_alpha
        self.multiple_latents = multiple_latents
        self.u_dim = u_dim
        self.w_dim = w_dim 
        if self.multiple_latents:
            assert self.u_dim is not None and self.w_dim is not None, "u_dim and w_dim must be specified for multiple_latents=True"
            assert self.u_dim + self.w_dim == self.z_dim, "u_dim + w_dim must equal z_dim"
            self.layer_sizes[-1] = self.u_dim
        self.encoder_layers = self.encoder_layers[:-1]
        self.enc_mean_layer = nn.Linear(
            self.layer_sizes[-2],
            self.layer_sizes[-1],
            bias=self.bias,
        )

        if not self.sparse:
            self.enc_logvar_layer = nn.Linear(
                self.layer_sizes[-2],
                self.layer_sizes[-1],
                bias=self.bias,
            )

        if self.multiple_latents:
            self.enc_mean_layer_private = nn.Linear(
                self.layer_sizes[-2],
                self.w_dim,
                bias=self.bias,
            )
            self.enc_logvar_layer_private = nn.Linear(
                self.layer_sizes[-2],
                self.w_dim,
                bias=self.bias,
            )

    def forward(self, x):
        h1 = x
        for it_layer, layer in enumerate(self.encoder_layers):
            h1 = layer(h1)
            if self.non_linear:
                h1 = F.relu(h1)

        if not self.sparse:
            mu = self.enc_mean_layer(h1)
            logvar = self.enc_logvar_layer(h1)
            if self.multiple_latents:
                mu_private = self.enc_mean_layer_private(h1)
                logvar_private = self.enc_logvar_layer_private(h1)
                return mu, logvar, mu_private, logvar_private
        else:
            mu = self.enc_mean_layer(h1)
            logvar = self.log_alpha + 2 * torch.log(torch.abs(mu) + 1e-8)

        return mu, logvar

class ConditionalVariationalEncoder(Encoder):
    """MLP Variational Conditional Encoder

    Args:
        y (list):  
        input_dim (list): Dimensionality of the input data.
        z_dim (int): Number of latent dimensions.
        hidden_layer_dim (list): Number of nodes per hidden layer.
        non_linear (bool): Whether to include a ReLU() function between layers.
        bias (bool): Whether to include a bias term in hidden layers.
        sparse (bool): Whether to enforce sparsity of the encoding distribution.
        log_alpha (float): Log of the dropout parameter.
        enc_dist (multiviewae.base.distributions.Normal, multiviewae.base.distributions.MultivariateNormal): Encoder distribution.
        num_cat (int): Number of categories of the labels.
        one_hot (bool): Whether to one-hot encode the labels.
        multiple_latents (bool, optional): Whether the model using a separate linear layers for shared and private latent spaces.
        u_dim (int, optional): Dimensionality of the shared latent space.
        w_dim (int, optional): Dimensionality of the private latent space.
    """
    def __init__(
        self,
        input_dim,
        z_dim,
        hidden_layer_dim,
        non_linear,
        bias,
        sparse,
        log_alpha,
        enc_dist,
        num_cat,
        one_hot,
        multiple_latents=False, 
        u_dim=None,
        w_dim=None
    ):
        super().__init__(input_dim=input_dim,
                        z_dim=z_dim,
                        hidden_layer_dim=hidden_layer_dim,
                        bias=bias,
                        non_linear=non_linear,
                        enc_dist=enc_dist)

        self.num_cat = num_cat
        self.sparse = sparse
        self.non_linear = non_linear
        self.log_alpha = log_alpha
        self.one_hot = one_hot
        self.multiple_latents = multiple_latents
        self.u_dim = u_dim
        self.w_dim = w_dim
        self.layer_sizes = [input_dim + num_cat] + self.hidden_layer_dim + [z_dim]
        if self.multiple_latents:
            assert self.u_dim is not None and self.w_dim is not None, "u_dim and w_dim must be specified for multiple_latents=True"
            assert self.u_dim + self.w_dim == self.z_dim, "u_dim + w_dim must equal z_dim"
            self.layer_sizes[-1] = self.u_dim
        lin_layers = [
            nn.Linear(dim0, dim1, bias=self.bias)
            for dim0, dim1 in zip(
                self.layer_sizes[:-1], self.layer_sizes[1:]
            )
        ]
        self.encoder_layers = nn.Sequential(*lin_layers)

        self.encoder_layers = self.encoder_layers[:-1]
        self.enc_mean_layer = nn.Linear(
            self.layer_sizes[-2],
            self.layer_sizes[-1],
            bias=self.bias,
        )

        if not self.sparse:
            self.enc_logvar_layer = nn.Linear(
                self.layer_sizes[-2],
                self.layer_sizes[-1],
                bias=self.bias,
            )

        if self.multiple_latents:
            self.enc_mean_layer_private = nn.Linear(
                self.layer_sizes[-2],
                self.w_dim,
                bias=self.bias,
            )
            self.enc_logvar_layer_private = nn.Linear(
                self.layer_sizes[-2],
                self.w_dim,
                bias=self.bias,
            )


    def set_labels(self, labels):
        self.labels = labels 

    def forward(self, x):
        if self.one_hot:
            c = F.one_hot(self.labels.long(), self.num_cat)
        else:
            c = self.labels

        x_cond = torch.hstack((x, c))   
        
        h1 = x_cond
        for it_layer, layer in enumerate(self.encoder_layers):
            h1 = layer(h1)
            if self.non_linear:
                h1 = F.relu(h1)

        if not self.sparse:
            mu = self.enc_mean_layer(h1)
            logvar = self.enc_logvar_layer(h1)
            if self.multiple_latents:
                mu_private = self.enc_mean_layer_private(h1)
                logvar_private = self.enc_logvar_layer_private(h1)
                return mu, logvar, mu_private, logvar_private
        else:
            mu = self.enc_mean_layer(h1)
            logvar = self.log_alpha + 2 * torch.log(torch.abs(mu) + 1e-8)

        return mu, logvar
    
class Decoder(nn.Module):
    """MLP Decoder
    
    Args:
        input_dim (list): Dimensionality of the input data.
        z_dim (int): Number of latent dimensions.
        hidden_layer_dim (list): Number of nodes per hidden layer. The layer order is reversed e.g. [100, 50, 5] becomes [5, 50, 100].
        non_linear (bool): Whether to include a ReLU() function between layers.
        bias (bool): Whether to include a bias term in hidden layers.
        dec_dist (multiviewae.base.distributions.Default, multiviewae.base.distributions.Bernoulli): Decoder distribution.
        init_logvar (int, float): Initial value for log variance of decoder. Unused in Decoder class.
    """
    def __init__(
        self,
        input_dim,
        z_dim,
        hidden_layer_dim,
        non_linear,
        bias,
        dec_dist,
        init_logvar=None
    ):
        super().__init__()

        self.input_size = input_dim
        self.z_dim = z_dim
        self.hidden_layer_dim = hidden_layer_dim
        self.bias = bias
        self.dec_dist = dec_dist
        self.non_linear = non_linear
        self.layer_sizes = [z_dim] + self.hidden_layer_dim + [input_dim]
        lin_layers = [
            nn.Linear(dim0, dim1, bias=self.bias)
            for dim0, dim1 in zip(
                self.layer_sizes[:-1], self.layer_sizes[1:]
            )
        ]

        self.decoder_layers = nn.Sequential(*lin_layers)

    def forward(self, z):
        x_rec = z
        for it_layer, layer in enumerate(self.decoder_layers[:-1]):
            x_rec = layer(x_rec)
            if self.non_linear:
                x_rec = F.relu(x_rec)
        x_rec = self.decoder_layers[-1](x_rec)
        x_rec = hydra.utils.instantiate(self.dec_dist, x=x_rec)
        return x_rec

class VariationalDecoder(Decoder):
    """MLP Variational Decoder

    Args:
        input_dim (list): Dimensionality of the input data.
        z_dim (int): Number of latent dimensions.
        hidden_layer_dim (list): Number of nodes per hidden layer. The layer order is reversed e.g. [100, 50, 5] becomes [5, 50, 100].
        non_linear (bool): Whether to include a ReLU() function between layers.
        bias (bool): Whether to include a bias term in hidden layers.
        init_logvar (int, float): Initial value for log variance of decoder.
        dec_dist (multiviewae.base.distributions.Normal, multiviewae.base.distributions.MultivariateNormal): Decoder distribution.
    """
    def __init__(
        self,
        input_dim,
        z_dim,
        hidden_layer_dim,
        bias,
        non_linear,
        init_logvar,
        dec_dist
    ):
        super().__init__(input_dim=input_dim,
                z_dim=z_dim,
                hidden_layer_dim=hidden_layer_dim,
                non_linear=non_linear,
                bias=bias, dec_dist=dec_dist)

        self.non_linear = non_linear
        self.init_logvar = init_logvar
        self.dec_dist = dec_dist

        self.decoder_layers = self.decoder_layers[:-1]
        self.dec_mean_layer = nn.Linear(
            self.layer_sizes[-2],
            self.layer_sizes[-1],
            bias=self.bias,
        )
        tmp_noise_par = torch.FloatTensor(1, self.input_size).fill_(
            self.init_logvar
        )
        self.logvar_out = Parameter(data=tmp_noise_par, requires_grad=True)

    def forward(self, z):
        x_rec = z
        for it_layer, layer in enumerate(self.decoder_layers):
            x_rec = layer(x_rec)
            if self.non_linear:
                x_rec = F.relu(x_rec)

        x_rec = self.dec_mean_layer(x_rec)
        x_rec = hydra.utils.instantiate(
            self.dec_dist, loc=x_rec, scale=torch.exp(0.5 * self.logvar_out)
        )
        return x_rec
    
class ConditionalVariationalDecoder(Decoder):
    """MLP Conditinal Variational Decoder

    Args:
        input_dim (list): Dimensionality of the input data.
        z_dim (int): Number of latent dimensions.
        hidden_layer_dim (list): Number of nodes per hidden layer. The layer order is reversed e.g. [100, 50, 5] becomes [5, 50, 100].
        non_linear (bool): Whether to include a ReLU() function between layers.
        bias (bool): Whether to include a bias term in hidden layers.
        init_logvar (int, float): Initial value for log variance of decoder.
        dec_dist (multiviewae.base.distributions.Normal, multiviewae.base.distributions.MultivariateNormal): Decoder distribution.
        num_cat (int): Number of categories of the labels.
        one_hot (bool): Whether to one-hot encode the labels.
    """
    def __init__(
        self,
        input_dim,
        z_dim,
        hidden_layer_dim,
        bias,
        non_linear,
        init_logvar,
        dec_dist, 
        num_cat,
        one_hot
    ):
        super().__init__(input_dim=input_dim,
                z_dim=z_dim,
                hidden_layer_dim=hidden_layer_dim,
                non_linear=non_linear,
                bias=bias, dec_dist=dec_dist)

        self.num_cat = num_cat
        self.non_linear = non_linear
        self.init_logvar = init_logvar
        self.dec_dist = dec_dist
        self.one_hot = one_hot

        self.layer_sizes = [z_dim + num_cat] + self.hidden_layer_dim + [input_dim]
        lin_layers = [
            nn.Linear(dim0, dim1, bias=self.bias)
            for dim0, dim1 in zip(
                self.layer_sizes[:-1], self.layer_sizes[1:]
            )
        ]

        self.decoder_layers = nn.Sequential(*lin_layers)

        self.decoder_layers = self.decoder_layers[:-1]
        self.dec_mean_layer = nn.Linear(
            self.layer_sizes[-2],
            self.layer_sizes[-1],
            bias=self.bias,
        )
        tmp_noise_par = torch.FloatTensor(1, self.input_size).fill_(
            self.init_logvar
        )
        self.logvar_out = Parameter(data=tmp_noise_par, requires_grad=True)

    def set_labels(self, labels): 
        self.labels = labels 

    def forward(self, z):
        if self.one_hot:
            c = F.one_hot(self.labels.long(), self.num_cat)
        else:
            c = self.labels
  
        if (len(z.size()) == 3 and len(c.size()) == 2): # NOTE: for mmvae which uses rsample() instead of sample()
            z_cond = torch.cat((z, c.repeat(z.size()[0],1,1)), dim=2)
        else:
            z_cond = torch.hstack((z, c))
        
        x_rec = z_cond
        for it_layer, layer in enumerate(self.decoder_layers):
            x_rec = layer(x_rec)
            if self.non_linear:
                x_rec = F.relu(x_rec)

        x_rec = self.dec_mean_layer(x_rec)
        x_rec = hydra.utils.instantiate(
            self.dec_dist, loc=x_rec, scale=torch.exp(0.5 * self.logvar_out)
        )
        return x_rec

class Discriminator(nn.Module):
    """MLP Discriminator
    
    Args:
        input_dim (list): Dimensionality of the input data.
        z_dim (int): Number of output dimensions.
        hidden_layer_dim (list): Number of nodes per hidden layer.
        non_linear (bool): Whether to include a ReLU() function between layers.
        bias (bool): Whether to include a bias term in hidden layers.
        dropout_threshold (float): Dropout threshold of layers.
        is_wasserstein (bool): Whether model employs a wasserstein loss.
    """
    def __init__(
        self,
        input_dim,
        output_dim,
        hidden_layer_dim,
        non_linear,
        bias,
        dropout_threshold,
        is_wasserstein
    ):
        super().__init__()
        self.bias = bias
        self.non_linear = non_linear
        self.dropout_threshold = dropout_threshold
        self.is_wasserstein = is_wasserstein

        self.layer_sizes = [input_dim] + hidden_layer_dim + [output_dim]

        lin_layers = [
            nn.Linear(dim0, dim1, bias=self.bias)
            for dim0, dim1 in zip(self.layer_sizes[:-1], self.layer_sizes[1:])
        ]
        self.linear_layers = nn.Sequential(*lin_layers)

    def forward(self, x):
        for it_layer, layer in enumerate(self.linear_layers):
            x = F.dropout(layer(x), self.dropout_threshold, training=self.training)
            if it_layer < len(self.linear_layers) - 1:
                if self.non_linear:
                    x = F.relu(x)
            else:
                if self.is_wasserstein:
                    return x
                elif self.layer_sizes[-1] > 1:
                    x = nn.Softmax(dim=-1)(x)
                else:
                    x = torch.sigmoid(x)
        return x



import torch
from torch.distributions import Normal, kl_divergence, Laplace, Bernoulli
from torch.distributions.multivariate_normal import MultivariateNormal
from torch.distributions.utils import broadcast_all
from torch.nn.functional import binary_cross_entropy
from .constants import EPS
import torch.nn.functional as F

def compute_log_alpha(mu, logvar):
    # clamp because dropout rate p in 0-99%, where p = alpha/(alpha+1)
    return (logvar - 2 * torch.log(torch.abs(mu) + 1e-8)).clamp(min=-8, max=8)

class Default():
    """Artificial distribution designed for data with unspecified distribution.
    Used so that log_likelihood and _sample methods can be called by model class.
    Args:
        x (list): List of input data.
    """
    def __init__(
        self,
        **kwargs,
    ):
        self.x = kwargs['x']


    def log_likelihood(self, x):
        """calculates the mean squared error between input data and reconstruction.

        Args:
            x (torch.Tensor): data reconstruction.

        Returns:
            torch.Tensor: Negative mean squared error.
        """
        logits, x = broadcast_all(self.x, x)
        return - (logits - x)**2

    def rsample(self):
        raise NotImplementedError

    def kl_divergence(self):
        raise NotImplementedError

    def sparse_kl_divergence(self):
        raise NotImplementedError

    def _sample(self, training=False, return_mean=True):
        return self.x

class Categorical():    
    """Artificial distribution designed for categorical data.
    Used so that log_likelihood and _sample methods can be called by model class.
    Args:
        x (list): List of input data.
    """
    def __init__(
        self,
        **kwargs,
    ):
        self.x = kwargs['x']

    def log_likelihood(self, x, eps=1e-6):
        """calculates the k-class cross entropy between input data and reconstruction. 
        Args:
            x (torch.Tensor): data reconstruction.

        Returns:
            torch.Tensor: Negative k-class cross entropy.
        """
        log_input = F.log_softmax(self.x + eps, dim=-1)
        ce = x * log_input
        return ce
    
    def rsample(self):
        raise NotImplementedError

    def kl_divergence(self):
        raise NotImplementedError

    def sparse_kl_divergence(self):
        raise NotImplementedError

    def _sample(self, training=False, return_mean=True):
        return self.x
          
class Normal(Normal):
    """Univariate normal distribution. Inherits from torch.distributions.Normal.

    Args:
        loc (int, torch.Tensor): Mean of distribution.
        scale (int, torch.Tensor): Standard deviation of distribution.
    """
    def __init__(
        self,
        **kwargs,
    ):
        self.loc = kwargs['loc']
        if 'logvar' in kwargs:
            self.logvar = kwargs['logvar']
            self.scale = kwargs['logvar'].mul(0.5).exp_()

        elif 'scale' in kwargs:
            self.scale = kwargs['scale']
            if not isinstance(self.scale, torch.Tensor):
                self.scale = torch.tensor(self.scale)
            self.logvar = 2 * torch.log(self.scale)
        super().__init__(loc=self.loc, scale=self.scale)

    @property
    def variance(self):
        return self.scale.pow(2)

    def kl_divergence(self, other):
        logvar0 = self.logvar
        mu0 = self.loc
        logvar1 = other.logvar
        mu1 = other.loc

        return -0.5 * (1 - logvar0.exp()/logvar1.exp() - (mu0-mu1).pow(2)/logvar1.exp() + logvar0 - logvar1)
    
    def sparse_kl_divergence(self):
        """
        Implementation from: https://github.com/senya-ashukha/variational-dropout-sparsifies-dnn/blob/master/KL%20approximation.ipynb

        """
        mu = self.loc
        logvar = torch.log(self.variance)
        log_alpha = compute_log_alpha(mu, logvar)
        k1, k2, k3 = 0.63576, 1.8732, 1.48695
        neg_KL = (
            k1 * torch.sigmoid(k2 + k3 * log_alpha)
            - 0.5 * torch.log1p(torch.exp(-log_alpha))
            - k1
        )
        return -neg_KL

    def log_likelihood(self, x):
        return self.log_prob(x)

    def _sample(self, *kwargs, training=False, return_mean=True):
        if training:
            return self.rsample(*kwargs)

        if return_mean:
                return self.loc
        return self.rsample()

class MultivariateNormal(MultivariateNormal):
    """Multivariate normal distribution with diagonal covariance matrix. Inherits from torch.distributions.multivariate_normal.MultivariateNormal.

    Args:
        loc (list, torch.Tensor): Mean of distribution.
        scale (int, torch.Tensor): Standard deviation of distribution.
    """
    def __init__(
            self,
            **kwargs
        ):

        self.loc = torch.as_tensor(kwargs['loc'])
        if 'logvar' in kwargs:
            self.logvar = torch.as_tensor(kwargs['logvar'])
            self.scale = torch.exp(0.5 * kwargs['logvar']) + EPS

        elif 'scale' in kwargs:
            self.scale = torch.as_tensor(kwargs['scale'])
            if not isinstance(self.scale, torch.Tensor):
                self.scale = torch.tensor(self.scale)
            self.logvar = 2 * torch.log(self.scale)
        
        #used when fitting encoder/decoder distribution or prior distribution with different mean and SD values
        self.covariance_matrix = torch.diag_embed(self.scale)

        super().__init__(loc=self.loc, covariance_matrix=self.covariance_matrix)

    @property
    def variance(self):
        return self.scale.pow(2)

    def kl_divergence(self, other):

        kl = kl_divergence(torch.distributions.multivariate_normal.MultivariateNormal( \
                        loc=self.loc, covariance_matrix=self.covariance_matrix), other)
        return torch.unsqueeze(kl,-1)

    def sparse_kl_divergence(self):
        mu = self.loc
        logvar = torch.log(self.variance)
        log_alpha = compute_log_alpha(mu, logvar)
        k1, k2, k3 = 0.63576, 1.8732, 1.48695
        neg_KL = (
            k1 * torch.sigmoid(k2 + k3 * log_alpha)
            - 0.5 * torch.log1p(torch.exp(-log_alpha))
            - k1
        )
        return -neg_KL

    def log_likelihood(self, x):
        ll = self.log_prob(x)
        return torch.unsqueeze(ll,-1)

    def _sample(self, *kwargs, training=False, return_mean=True):
        if training:
            return self.rsample(*kwargs)
        if return_mean:
                return self.loc
        return self.rsample()

class Bernoulli(Bernoulli):
    """Bernoulli distribution. Inherits from torch.distributions.Bernoulli.
    Args:
        x (list): List of input data.
    """
    def __init__(
        self,
        **kwargs,
    ):
        x = kwargs['x']
        super().__init__(logits=x)

    def log_likelihood(self, target):   
        return self.log_prob(target)

    def rsample(self):
        raise NotImplementedError

    def kl_divergence(self):
        raise NotImplementedError

    def sparse_kl_divergence(self):
        raise NotImplementedError

    def _sample(self, training=False, return_mean=True):
        return self.sample()
    
class ApproxBernoulli():
    """Artificial distribution designed for (approximately) Bernoulli distributed data. 
    The data isn't restricted to bernoulli distribution, this class is designed as a wrapper for the log_likelihood() method which is required for the multiview models.

    Args:
        x (list): List of input data.
    """
    def __init__(
        self,
        **kwargs,
    ):
        self.x = kwargs['x']

    def log_likelihood(self, target):   
        x, target = broadcast_all(self.x, target)
        x = torch.sigmoid(x)
        bce = binary_cross_entropy(x, target, reduction='none')
        return -bce

    def rsample(self):
        raise NotImplementedError

    def kl_divergence(self):
        raise NotImplementedError

    def sparse_kl_divergence(self):
        raise NotImplementedError

    def _sample(self, training=False, return_mean=True):
        return torch.sigmoid(self.x)

class Laplace(Laplace):
    """Laplace distribution. Inherits from torch.distributions.Laplace.

    Args:
        loc (list, torch.Tensor): Mean of distribution.
        scale (int, torch.Tensor): Standard deviation of distribution.
    """
    def __init__(
            self,
            **kwargs
        ):
        if 'loc' in kwargs:
            self.loc = torch.as_tensor(kwargs['loc'])
        elif 'x' in kwargs:
            self.loc = torch.as_tensor(kwargs['x'])
 
        if 'logvar' in kwargs:
            self.logvar = torch.as_tensor(kwargs['logvar'])
            if 'with_softmax' in kwargs and kwargs['with_softmax']:
                self.scale = (F.softmax(self.logvar, dim=-1) * self.logvar.size(-1) + EPS).to(self.loc.device)
            else:   
                self.scale = torch.exp(0.5 * self.logvar) + EPS
        elif 'scale' in kwargs:
            self.scale = torch.as_tensor(kwargs['scale'])
        else:
            self.scale = torch.tensor(0.75).to(self.loc.device)

        super().__init__(loc=self.loc, scale=self.scale)

    def kl_divergence(self, other):
        
        kl = kl_divergence(torch.distributions.laplace.Laplace( \
                        loc=self.loc, scale=self.scale), other) #check this works
        return torch.unsqueeze(kl,-1)
    
    def sparse_kl_divergence(self):
        raise NotImplementedError
        
    def log_likelihood(self, x):
        return self.log_prob(x)

    def _sample(self, *kwargs, training=False, return_mean=True):
        if training:
            return self.rsample(*kwargs)

        if return_mean:
                return self.loc
        return self.rsample()



import torch
import hydra

from ..base.constants import MODEL_WEIGHTEDDMVAE
from ..base.base_model import BaseModelVAE
from ..base.representations import weightedProductOfExperts

class weighted_DMVAE(BaseModelVAE):
    r"""
    Variant of Disentangled multi-modal variational autoencoder (DMVAE) using weighted Product-of-Experts for joint encoding distribution.

    Args:
        cfg (str): Path to configuration file. Model specific parameters in addition to default parameters:
            
            - model._lambda (list, optional): Log likelihood weighting term for each modality.
            - model.s_dim (int): Number of private latent dimensions.
            - model.beta (int, float): KL divergence weighting term.
            - encoder.default._target_ (multiviewae.architectures.mlp.VariationalEncoder): Type of encoder class to use.
            - encoder.default.enc_dist._target_ (multiviewae.base.distributions.Normal, multiviewae.base.distributions.MultivariateNormal): Encoding distribution.
            - decoder.default._target_ (multiviewae.architectures.mlp.VariationalDecoder): Type of decoder class to use.
            - decoder.default.init_logvar (int, float): Initial value for log variance of decoder.
            - decoder.default.dec_dist._target_ (multiviewae.base.distributions.Normal, multiviewae.base.distributions.MultivariateNormal): Decoding distribution.
        
        input_dim (list): Dimensionality of the input data.
        z_dim (int): Number of latent dimensions.

    References
    ----------
    M. Lee and V. Pavlovic, "Private-Shared Disentangled Multimodal VAE for Learning of Latent Representations," 2021 IEEE/CVF Conference on Computer Vision and 
    Pattern Recognition Workshops (CVPRW), Nashville, TN, USA, 2021, pp. 1692-1700, doi: 10.1109/CVPRW53098.2021.00185.

    """

    def __init__(
        self,
        cfg = None,
        input_dim = None,
        z_dim = None
    ):
        super().__init__(model_name=MODEL_WEIGHTEDDMVAE,
                        cfg=cfg,
                        input_dim=input_dim,
                        z_dim=z_dim)

        self.join_z = weightedProductOfExperts()
        #if lambda is none, set to 1 for all modalities
        if not hasattr(self, '_lambda'):
            self._lambda = [1 for i in range(self.n_views)]
        else:
            assert len(self._lambda) == self.n_views, "Number of lambda weightings must be equal to number of views"

        tmp_weight = torch.FloatTensor(len(input_dim), self.z_dim - self.s_dim).fill_(1/len(input_dim))
        self.poe_weight = torch.nn.Parameter(data=tmp_weight, requires_grad=True)
    def encode(self, x):
        r"""Forward pass through encoder networks.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            qc_x (list): Single element list containing the PoE shared encoding distribution.
            qcs_xs (list): list containing encoding distributions for shared latent dimensions for each view.
            qs_xs (list): list of encoding distributions for private latents.
            qscs_xs (list): nested list containing combined PoE shared and private latents.
            qscss_xs (list): nested list containing combined shared latents from each modality and private latents for same and cross view reconstruction.
            
        """
        qs_xs = []
        qcs_xs = []
        qscs_xs = []
        qscss_xs = []
        mu_s = []
        logvar_s = []
        mu_c = []
        logvar_c = []
        for i in range(self.n_views):
            mu, logvar = self.encoders[i](x[i])
            mu_s.append(mu[:,:self.s_dim])
            logvar_s.append(logvar[:,:self.s_dim])
            mu_c.append(mu[:,self.s_dim:])
            logvar_c.append(logvar[:,self.s_dim:])

            qs_x = hydra.utils.instantiate(
            eval(f"self.cfg.encoder.enc{i}.enc_dist"), loc=mu[:,:self.s_dim], logvar=logvar[:,:self.s_dim]
            )
            qs_xs.append(qs_x)
            qc_x = hydra.utils.instantiate(
            eval(f"self.cfg.encoder.enc{i}.enc_dist"), loc=mu[:,self.s_dim:], logvar=logvar[:,self.s_dim:]
            )
            qcs_xs.append(qc_x)

        for i in range(self.n_views):
            qscs_x_ = []
            for j in range(self.n_views):
                mu_sc = torch.cat((mu_s[i], mu_c[j]), 1) #when i /= j for recon of view from shared latent of other view
                logvar_sc = torch.cat((logvar_s[i], logvar_c[j]), 1)
                qscs_x = hydra.utils.instantiate(
                eval(f"self.cfg.encoder.enc{i}.enc_dist"), loc=mu_sc, logvar=logvar_sc
                )
                qscs_x_.append(qscs_x)
            qscss_xs.append(qscs_x_)

        mu_c = torch.stack(mu_c)
        logvar_c = torch.stack(logvar_c)
    
        mu_c, logvar_c = self.join_z(mu_c, logvar_c, self.poe_weight)
        qc_x = hydra.utils.instantiate(
            self.cfg.encoder.default.enc_dist, loc=mu_c, logvar=logvar_c
        )
        with torch.no_grad():
            self.poe_weight = self.poe_weight.clamp_(0, +1)

        for i in range(self.n_views):   
            mu_sc = torch.cat((mu_s[i], mu_c), 1)
            logvar_sc = torch.cat((logvar_s[i], logvar_c), 1)
            qsc_x = hydra.utils.instantiate( 
            eval(f"self.cfg.encoder.enc{i}.enc_dist"), loc=mu_sc, logvar=logvar_sc
            )
            qscs_xs.append(qsc_x)

        if self._training:
            return [[qc_x], qcs_xs, qs_xs, qscs_xs, qscss_xs]
        return qscs_xs


    def decode(self, qz_x):
        r"""Forward pass of joint latent dimensions through decoder networks.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            (list): A nested list of decoding distributions, px_zs. The outer list has a single element indicating the shared latent dimensions. 
            The inner list is a n_view element list with the position in the list indicating the decoder index.
        """  
        px_zs = []
        for i in range(self.n_views):
            px_z = self.decoders[i](qz_x[i]._sample(training=self._training, return_mean=self.return_mean))
            px_zs.append(px_z)
        return [px_zs]

    def decode_separate(self, qz_xs):
        r"""Forward pass through decoder networks. Each shared latent is passed through all of the decoders with the private latents from the same view.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            (list): A nested list of decoding distributions, px_zs. The outer list has a n_view element list with position in the list indicating the decoder index. 
            The inner list is a n_view element list with the position in the list indicating latent dimensions index. NOTE: This is the reverse to other models.
        """  
        px_zs = []
        for i in range(self.n_views):
            px_z = [
                self.decoders[i](qz_xs[j][i]._sample(training=self._training, return_mean=self.return_mean)) #TODO: check this is right
                for j in range(self.n_views)
            ]
            px_zs.append(px_z)
        return px_zs 

    def forward(self, x):
        r"""Apply encode and decode methods to input data to generate the latent dimensions and data reconstructions. 
        
        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            fwd_rtn (dict): dictionary containing encoding and decoding distributions.
        """
        qc_x, qcs_xs, qs_xs, qscs_xs, qscss_xs = self.encode(x)
        px_zs = self.decode(qscs_xs)
        pxs_zs = self.decode_separate(qscss_xs)
        fwd_rtn = {"px_zs": px_zs, "pxs_zs": pxs_zs, "qcs_xs": qcs_xs, "qs_xs": qs_xs, "qc_x": qc_x}

        return fwd_rtn


    def calc_kl_joint_latent(self, qz_x, qs_xs):
        r"""Calculate KL-divergence loss for the first terms in Equation 3.

        Args:
            qz_x (list): Single element list containing joint encoding distribution.
            qs_xs (list): list of encoding distributions for private latent dimensions for each view.

        Returns:
            (torch.Tensor): KL-divergence loss.
        """
        kl = 0
        for i in range(self.n_views):
            kl += qs_xs[i].kl_divergence(self.prior).mean(0).sum()
            kl += qz_x[0].kl_divergence(self.prior).mean(0).sum()

        return kl*self.beta
    
    def calc_kl_separate_latent(self, qcs_xs, qs_xs):
        r"""Calculate KL-divergence loss for the second terms in Equation 3.

        Args:
            qcs_x (list): list of the shared encoding distributions calculated from each view.
            qs_xs (list): list of encoding distributions for private latent dimensions for each view.

        Returns:
            (torch.Tensor): KL-divergence loss.
        """
        kl = 0
        for i in range(self.n_views):
            for j in range(self.n_views):
                kl += qcs_xs[j].kl_divergence(self.prior).mean(0).sum() 
                kl += qs_xs[i].kl_divergence(self.prior).mean(0).sum()
        return kl*self.beta
    
    def calc_ll_joint(self, x, px_zs):
        r"""Calculate log-likelihood loss from the joint encoding distribution for each modality. 

        Args:
            x (list): list of input data of type torch.Tensor.
            px_zs (list): list of decoding distributions. 

        Returns:
            ll (torch.Tensor): Log-likelihood loss.
        """
        ll = 0
        for i in range(self.n_views):
            ll += px_zs[0][i].log_likelihood(x[i]).mean(0).sum() #first index is latent, second index is view
        return ll

    def calc_ll_separate(self, x, pxs_zs):
        r"""Calculate cross-modal and self-reconstrution log-likelihood loss from the shared encoding distribution for each modality and private latents.

        Args:
            x (list): list of input data of type torch.Tensor.
            pxs_zs (list): nested list of decoding distributons. NOTE: The ordering of decoding distribution is the reverse compared to other models.

        Returns:
            ll (torch.Tensor): Log-likelihood loss.
        """
        ll = 0
        for i in range(self.n_views):
            for j in range(self.n_views):
                ll += self._lambda[i]*pxs_zs[i][j].log_likelihood(x[i]).mean(0).sum() #first index is view, second index is latent
        return ll    
               
    def loss_function(self, x, fwd_rtn):
        r"""Calculate DMVAE loss.
        
        Args:
            x (list): list of input data of type torch.Tensor.
            fwd_rtn (dict): dictionary containing encoding and decoding distributions.

        Returns:
            losses (dict): dictionary containing each element of the DMVAE loss.
        """
        px_zs = fwd_rtn["px_zs"]
        ll = self.calc_ll_joint(x, px_zs)

        qc_x = fwd_rtn["qc_x"]
        qcs_xs = fwd_rtn["qcs_xs"]
        qs_xs = fwd_rtn["qs_xs"]
        pxs_zs = fwd_rtn["pxs_zs"] 

        kl = self.calc_kl_joint_latent(qc_x, qs_xs) 
        kl += self.calc_kl_separate_latent(qcs_xs, qs_xs) 
        ll += self.calc_ll_separate(x, pxs_zs) 
        
        total = kl - ll
        losses = {"loss": total, "kl": kl, "ll": ll}
        return losses

    def configure_optimizers(self):
        optimizers = [
            torch.optim.Adam(
                list(self.encoders[i].parameters())
                + list(self.decoders[i].parameters()), 
                lr=self.learning_rate,
            )
            for i in range(self.n_views)
        ]
        optimizers.append(torch.optim.Adam([self.poe_weight],
                lr=self.learning_rate
            ))
        return optimizers


from ..base.constants import MODEL_DCCAE, EPS
from ..base.base_model import BaseModelAE
import torch

class DCCAE(BaseModelAE):
    r"""Deep Canonically Correlated Autoencoder (DCCAE).
    CCA implementation adapted from: https://github.com/jameschapman19/cca_zoo

    Args:
        cfg (str): Path to configuration file. Model specific parameters in addition to default parameters:
            
            - model._lambda (int, float): Reconstruction weighting term

        input_dim (list): Dimensionality of the input data.
        z_dim (int): Number of latent dimensions.

    References
    ----------
    Wang, Weiran & Arora, Raman & Livescu, Karen & Bilmes, Jeff. (2016). On Deep Multi-View Representation Learning: Objectives and Optimization.

    """
    def __init__(
        self,
        cfg = None,
        input_dim = None,
        z_dim = None
    ):
        super().__init__(model_name=MODEL_DCCAE,
                        cfg=cfg,
                        input_dim=input_dim,
                        z_dim=z_dim)

    def encode(self, x):
        r"""Forward pass through encoder networks.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            z (list): list of latent dimensions for each view of type torch.Tensor.
        """   
        z = []
        for i in range(self.n_views):
            z_ = self.encoders[i](x[i])
            z.append(z_)
        return z

    def decode(self, z):
        r"""Forward pass through decoder networks. Each latent is passed through all of the decoders.

        Args:
            z (list): list of latent dimensions for each view of type torch.Tensor.

        Returns:
            x_recon (list): list of data reconstructions.
        """
        x_recon = []
        for i in range(self.n_views):
            temp_recon = self.decoders[i](z[i])
            x_recon.append(temp_recon)
        return [x_recon]

    def cca(self, zs):
        r"""CCA loss calculation.
        Adapted from: https://github.com/jameschapman19/cca_zoo/blob/main/cca_zoo/deep/_discriminative/_dmcca.py

        Args:
            z (list): list of latent dimensions for each view of type torch.Tensor.

        Returns:
            cca_loss (torch.Tensor): CCA loss.
        """

        zs = [
            z - z.mean(dim=0)
            for z in zs
        ]
        all_views = torch.cat(zs, dim=1)

        #Calculate cross-covariance matrix
        C = torch.cov(all_views.T)
        C = C - torch.block_diag(
            *[torch.cov(z.T) for z in zs]
        )
        C =  C / len(zs)

        #Calculate block covariance matrix
        D = torch.block_diag(
            *[
                (1 - EPS) * torch.cov(z.T)
                + EPS
                * torch.eye(z.shape[1], device=z.device)
                for z in zs
            ]
        )
        D = D / len(zs)

        C += D

        U, S, V = torch.svd(D)
        # Enforce positive definite by taking a torch max() with EPS
        S = torch.clamp(S, min=EPS)
        # Calculate inverse square-root
        inv_sqrt_S = torch.diag_embed(torch.pow(S, -0.5))
        # Calculate inverse square-root matrix
        R = torch.matmul(torch.matmul(U, inv_sqrt_S), V.transpose(-1, -2))

        C_whitened = R @ C @ R.T
        eigvals = torch.linalg.eigvalsh(C_whitened)
        idx = torch.argsort(eigvals, descending=True)
        eigvals = eigvals[idx[:self.z_dim]]
        eigvals = torch.nn.LeakyReLU()(eigvals[torch.gt(eigvals, 0)])
        corr = eigvals.sum()

        return -corr
    
    def forward(self, x):
        r"""Apply encode and decode methods to input data to generate latent dimensions and data reconstructions.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            fwd_rtn (dict): dictionary containing list of data reconstructions (x_recon) and latent dimensions (z).
        """
        z = self.encode(x)
        x_recon = self.decode(z)
        fwd_rtn = {"x_recon": x_recon, "z": z}
        return fwd_rtn

    def loss_function(self, x, fwd_rtn):    
        r"""Calculate reconstruction loss.
        
        Args:
            x (list): list of input data of type torch.Tensor.
            fwd_rtn (dict): dictionary containing list of data reconstructions (x_recon) and latent dimensions (z).

        Returns:
            losses (dict): dictionary containing reconstruction loss.
        """
        x_recon = fwd_rtn["x_recon"]
        z = fwd_rtn["z"]
        recon = 0
        for i in range(self.n_views):
            recon += - x_recon[0][i].log_likelihood(x[i]).mean(0).sum() #first index is latent, second index is view

        cca_loss = self.cca(z)        
        total_loss = self._lambda * recon + cca_loss
        losses = {"loss": total_loss, "recon_loss": recon, "cca_loss": cca_loss}
        return losses


import torch
import hydra
import numpy as np
from ..base.constants import MODEL_MVTCAE
from ..base.base_model import BaseModelVAE
from ..base.representations import ProductOfExperts

class mvtCAE(BaseModelVAE):
    r"""
    Multi-View Total Correlation Auto-Encoder (MVTCAE).

    Code is based on: https://github.com/gr8joo/MVTCAE

    NOTE: This implementation currently only caters for a PoE posterior distribution. MoE and MoPoE posteriors will be included in further work.

    Args:
        cfg (str): Path to configuration file. Model specific parameters in addition to default parameters:
            
            - model.beta (int, float): KL divergence weighting term.
            - model.alpha (int, float): Log likelihood, Conditional VIB and VIB weighting term.
            - encoder.default._target_ (multiviewae.architectures.mlp.VariationalEncoder): Type of encoder class to use.
            - encoder.default.enc_dist._target_ (multiviewae.base.distributions.Normal, multiviewae.base.distributions.MultivariateNormal): Encoding distribution.
            - decoder.default._target_ (multiviewae.architectures.mlp.VariationalDecoder): Type of decoder class to use.
            - decoder.default.init_logvar(int, float): Initial value for log variance of decoder.
            - decoder.default.dec_dist._target_ (multiviewae.base.distributions.Normal, multiviewae.base.distributions.MultivariateNormal): Decoding distribution.
    
        input_dim (list): Dimensionality of the input data.
        z_dim (int): Number of latent dimensions.

    References
    ----------
    Hwang, HyeongJoo and Kim, Geon-Hyeong and Hong, Seunghoon and Kim, Kee-Eung. Multi-View Representation Learning via Total Correlation Objective. 2021. NeurIPS

    """

    def __init__(
        self,
        cfg = None,
        input_dim = None,
        z_dim = None
    ):

        super().__init__(model_name=MODEL_MVTCAE,
                        cfg=cfg,
                        input_dim=input_dim,
                        z_dim=z_dim)

    def encode(self, x):
        r"""Forward pass through encoder networks.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            Returns the separate and/or joint encoding distributions depending on whether the model is in the training stage: 
            qz_xs (list): list containing separate encoding distributions.
            qz_x (list): Single element list containing PoE joint encoding distribution.
        """
     
        if self._training:
            qz_xs = []
            mu = []
            logvar = []
            for i in range(self.n_views):
                mu_, logvar_ = self.encoders[i](x[i])
                mu.append(mu_)
                logvar.append(logvar_)
                qz_x_ = hydra.utils.instantiate(
                    eval(f"self.cfg.encoder.enc{i}.enc_dist"), loc=mu_, logvar=logvar_
                )
                qz_xs.append(qz_x_)

            mu = torch.stack(mu)
            logvar = torch.stack(logvar)
            mu, logvar = ProductOfExperts()(mu, logvar)
            qz_x = hydra.utils.instantiate( 
                self.cfg.encoder.default.enc_dist, loc=mu, logvar=logvar
            )
            return [qz_x], qz_xs
        else:
            mu = []
            logvar = []
            for i in range(self.n_views):
                mu_, logvar_ = self.encoders[i](x[i])
                mu.append(mu_)
                logvar.append(logvar_)

            mu = torch.stack(mu)
            logvar = torch.stack(logvar)
            mu, logvar = ProductOfExperts()(mu, logvar)
            qz_x = hydra.utils.instantiate( 
                self.cfg.encoder.default.enc_dist, loc=mu, logvar=logvar
            )
            qz_x = [qz_x]
            return qz_x
        
    def encode_subset(self, x, subset):
        r"""Forward pass through encoder networks for a subset of modalities.

        Args:
            x (list): list of input data of type torch.Tensor.
            subset (list): list of modalities to encode.

        Returns:
            Returns either the joint or separate encoding distributions depending on whether the model is in the training stage: 
            qz_xs (list): list containing separate encoding distributions.
            qz_x (list): Single element list containing PoE joint encoding distribution.
        """
        mu = []
        logvar = []
        for i in subset:
            mu_, logvar_ = self.encoders[i](x[i])
            mu.append(mu_)
            logvar.append(logvar_)

        mu = torch.stack(mu)
        logvar = torch.stack(logvar)
        mu, logvar = ProductOfExperts()(mu, logvar)
        qz_x = hydra.utils.instantiate( 
            self.cfg.encoder.default.enc_dist, loc=mu, logvar=logvar
        )
        return [qz_x]
    
    def decode(self, qz_x):
        r"""Forward pass of joint latent dimensions through decoder networks.

        Args:
            qz_x (list): list of joint encoding distribution.

        Returns:
            (list): A nested list of decoding distributions, px_zs. The outer list has a single element indicating the shared latent dimensions. 
            The inner list is a n_view element list with the position in the list indicating the decoder index.
        """  
        px_zs = []
        for i in range(self.n_views):
            px_z = self.decoders[i](qz_x[0]._sample(training=self._training, return_mean=self.return_mean))
            px_zs.append(px_z)
        return [px_zs]

    def decode_subset(self, qz_x, subset):
        r"""Forward pass of joint latent dimensions through decoder networks for a subset of modalities.
        """
        px_zs = []
        for i in subset:
            px_z = self.decoders[i](qz_x[0]._sample(training=self._training, return_mean=self.return_mean))
            px_zs.append(px_z)
        return [px_zs]
    
    def forward(self, x):
        r"""Apply encode and decode methods to input data to generate the joint latent dimensions and data reconstructions. 
        
        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            fwd_rtn (dict): dictionary containing encoding and decoding distributions.
        """
        qz_x, qz_xs = self.encode(x)
        px_zs = self.decode(qz_x)
        fwd_rtn = {"px_zs": px_zs, "qz_xs": qz_xs, "qz_x": qz_x}
        return fwd_rtn

    def loss_function(self, x, fwd_rtn):
        r"""Calculate MVTCAE loss.
        
        Args:
            x (list): list of input data of type torch.Tensor.
            fwd_rtn (dict): dictionary containing encoding and decoding distributions.

        Returns:
            losses (dict): dictionary containing each element of the MVTCAE loss.
        """
        px_zs = fwd_rtn["px_zs"]
        qz_xs = fwd_rtn["qz_xs"]
        qz_x = fwd_rtn["qz_x"]

        rec_weight = (self.n_views - self.alpha) / self.n_views
        cvib_weight = self.alpha / self.n_views
        vib_weight = 1 - self.alpha

        grp_kl = self.calc_kl_groupwise(qz_x)
        cvib_kl = self.calc_kl_cvib(qz_x, qz_xs)
        ll = self.calc_ll(x, px_zs)

        kld_weighted = cvib_weight * cvib_kl + vib_weight * grp_kl
        total = -rec_weight * ll + self.beta * kld_weighted

        losses = {"loss": total, "kl_cvib": cvib_kl, "kl_grp": grp_kl, "ll": ll}
        return losses

    def calc_kl_cvib(self, qz_x, qz_xs):
        r"""Calculate KL-divergence between PoE joint encoding distribution and the encoding distribution for each view.

        Args:
            qz_xs (list): list of encoding distributions of each view.

        Returns:
            kl (torch.Tensor): KL-divergence loss.
        """
        kl = 0
        for i in range(self.n_views):
            kl += qz_x[0].kl_divergence(qz_xs[i]).mean(0).sum()
        return kl

    def calc_kl_groupwise(self, qz_x):
        r"""Calculate KL-divergence between the PoE joint encoding distribution and the prior distribution.

        Args:
            qz_xs (list): list of encoding distributions of each view.

        Returns:
            kl (torch.Tensor): KL-divergence loss.
        """
        return qz_x[0].kl_divergence(self.prior).mean(0).sum()
      

    def calc_ll(self, x, px_zs):
        r"""Calculate log-likelihood loss.

        Args:
            x (list): list of input data of type torch.Tensor.
            px_zs (list): list of decoding distributions.

        Returns:
            ll (torch.Tensor): Log-likelihood loss.
        """
        ll = 0
        for i in range(self.n_views):
            ll += px_zs[0][i].log_likelihood(x[i]).mean(0).sum() #first index is latent, second index is view
        return ll



from .ae import *
from .maae import *
from .mwae import *
from .mcvae import *
from .mvae import *
from .mmvae import *
from .mvtcae import *
from .dvcca import *
from .me_mvae import *
from .jmvae import *
from .mopoevae import *
from .mmJSD import *
from .weighted_mvae import *
from .DMVAE import *
from .weighted_DMVAE import *
from .mmvaeplus import *
from .dccae import *
__all__ = ["AE", "mAAE", "mWAE", "mcVAE", "mVAE", "JMVAE", "me_mVAE", "mmVAE", "mvtCAE", "DVCCA", "MoPoEVAE", "mmJSD", "weighted_mVAE",
         "DMVAE", "weighted_DMVAE", "mmVAEPlus", "DCCAE"]
classes = __all__ 




import os
import numpy as np
import hydra
import re
import collections.abc
import omegaconf 

import torch
import pytorch_lightning as pl

from os.path import join, isdir
from datetime import datetime
from abc import ABC, abstractmethod
from hydra import compose, initialize, initialize_config_dir
from schema import Schema, SchemaError, And, Or

from torch.utils.data.dataloader import DataLoader
from omegaconf import OmegaConf, open_dict

from .constants import *
from .validation import config_schema
from .exceptions import *
from .validation import SUPPORTED_DATASETS
from ..architectures.mlp import ConditionalVariationalEncoder, ConditionalVariationalDecoder

def update_dict(d, u, l):
    for k, v in u.items():
        if k in l:
            if isinstance(v, collections.abc.Mapping):
                d[k] = update_dict(d.get(k, {}), v, l=v.keys())
            else:
                d[k] = v
    return d

class BaseModelAE(ABC, pl.LightningModule):
    """Base class for autoencoder models.
    Args:
        model_name (str): Type of autoencoder model.
        cfg (str): Path to configuration file.
        input_dim (list): Dimensionality of the input data.
        z_dim (int): Number of latent dimensions.
    """
    is_variational = False

    @abstractmethod
    def __init__(
        self,
        model_name = None,
        cfg = None,
        input_dim = None,
        z_dim = None
    ):

        assert (model_name is not None and model_name in MODELS), \
        "Model name is invalid"  # have to choose which model always

        try:
            # check input_dim
            Schema(And(list, lambda x: len(x) > 0), error="input_dim should not be empty").validate(input_dim)
            Schema([Or(int, tuple)], error="input_dim should be a list of int/tuple").validate(input_dim)
            for d in input_dim:
                if isinstance(d, int):
                    Schema(lambda x: x > 0, error="each dim should be > 0").validate(d)
                else:
                    Schema(lambda x: all(a > 0 for a in x), error="each dim should be > 0").validate(d)
                    # Schema(lambda x: len(x) in [1,2,3], error="each dim should be 1D or 2D").validate(d)

            # check z_dim
            if z_dim is not None:
                Schema(And(int, lambda x: x > 0), error="z_dim must be > 0").validate(z_dim)
        except SchemaError as se:
            raise ConfigError(se) from None

        super().__init__()
        self.model_name = model_name
        self.input_dim = input_dim
        self.n_views = len(self.input_dim)
        self.z_dim = z_dim

        with initialize(version_base=None, config_path="../configs"):
            def_cfg = compose(
                            config_name="default",
                            return_hydra_config=True,
                            overrides=[f"model_type={self.model_name}.yaml"]
                        )

        user_cfg = None
        if cfg is not None: # user overrides default config
            if os.path.isabs(cfg):
                cfgdir, cfg_file = os.path.split(cfg)
                with initialize_config_dir(version_base=None, config_dir=cfgdir):
                    user_cfg = compose(
                                config_name=cfg_file,
                                return_hydra_config=True
                            )
            else:
                workdir = os.getcwd()
                with initialize_config_dir(version_base=None, config_dir=workdir):
                    user_cfg = compose(
                                config_name=cfg,
                                return_hydra_config=True
                            )
                    
        self.__initcfg(def_cfg, user_cfg)
        self.save_hyperparameters()

    ################################            public methods
    def fit(self, *data, labels=None, max_epochs=None, batch_size=None, cfg=None):
        if cfg is not None:
            new_cfg = omegaconf.OmegaConf.load(cfg)
            self.__initcfg(self.cfg, new_cfg, at_fit=True)
        data = list(data)

        if not self.is_index_ds:
            if not all(data_.shape[0] == data[0].shape[0] for data_ in data):
                raise InputError('All modalities must have the same number of entries')

            if not (len(data) == self.n_views):
                raise InputError("number of modalities must be equal to number of views")

            for i in range(self.n_views):
                data_dim = data[i].shape[1:]
                if len(data_dim) == 1:
                    data_dim = data_dim[0]
                if not (data_dim == self.input_dim[i]):
                    raise InputError("modality's shape must be equal to corresponding input_dim's shape")

        if any([isinstance(enc, ConditionalVariationalEncoder) for enc in self.encoders]) and labels is None: 
            raise InputError("no labels given for Conditional VAE")

        if any([isinstance(dec, ConditionalVariationalDecoder) for dec in self.decoders]) and labels is None: 
            raise InputError("no labels given for Conditional VAE")

        self._training = True
        if max_epochs is not None:
            self.max_epochs = max_epochs
            self.cfg.trainer.max_epochs = max_epochs
        else:
            self.max_epochs = self.cfg.trainer.max_epochs

        if batch_size is not None:
            self.batch_size = batch_size
            self.cfg.datamodule.batch_size = batch_size
        else:
            self.batch_size = self.cfg.datamodule.batch_size

        callbacks = []
        if self.cfg.datamodule.is_validate:
            for _, cb_conf in self.cfg.callbacks.items():
                callbacks.append(hydra.utils.instantiate(cb_conf))

        logger = hydra.utils.instantiate(self.cfg.logger)

        py_trainer = hydra.utils.instantiate(
            self.cfg.trainer, callbacks=callbacks, logger=logger,
        )
        datamodule = hydra.utils.instantiate(
           self.cfg.datamodule, data=data, n_views=self.n_views, labels=labels, _convert_="all", _recursive_=False
        )

        py_trainer.fit(self, datamodule)
        

    def predict_latents(self, *data, input_modalities=None, labels=None, batch_size=None):
        return self.__predict(*data, input_modalities=input_modalities, labels=labels, batch_size=batch_size)

    def predict_reconstruction(self, *data, input_modalities=None, output_modalities=None, labels=None, batch_size=None):
        return self.__predict(*data, input_modalities=input_modalities, output_modalities=output_modalities, labels=labels, batch_size=batch_size, is_recon=True)

    def print_config(self, cfg=None, keys=None):
        if cfg is None:
            cfg = self.cfg

        if keys is not None:
            print(f"{'model_name'}:\n  {cfg['model_name']}")
            for k in keys:
                if k in CONFIG_KEYS:
                    if cfg[k] is not None:
                        str = (OmegaConf.to_yaml(cfg[k])).replace("\n", "\n  ")
                    else:
                        str = "null\n"
                    print(f"{k}:\n  {str}")
        else:
            self.print_config(cfg=cfg, keys=CONFIG_KEYS)

    def save_config(self, keys=None):
        run_time = datetime.now().strftime("%Y-%m-%d_%H%M%S")
        save_cfg = {}
        if keys is not None:
            for k in keys:
                if k in CONFIG_KEYS:
                    save_cfg[k] = self.cfg[k]
            OmegaConf.save(save_cfg, join(self.cfg.out_dir, 'config_{0}.yaml'.format(run_time)))
        else:
            self.save_config(keys=CONFIG_KEYS)

    def create_folder(self, dir_path):
        check_folder = isdir(dir_path)
        if not check_folder:
            os.makedirs(dir_path)

    ################################            abstract methods

    @abstractmethod
    def encode(self, x):
        raise NotImplementedError()

    @abstractmethod
    def decode(self, z):
        raise NotImplementedError()

    @abstractmethod
    def loss_function(self, x, fwd_rtn):
        raise NotImplementedError()

    ################################            LightningModule methods
    @abstractmethod
    def forward(self, x):
         raise NotImplementedError()

    def training_step(self, batch, batch_idx, optimizer_idx=None):
        return self.__step(batch, batch_idx, stage="train")

    def validation_step(self, batch, batch_idx):
        return self.__step(batch, batch_idx, stage="val")

    def on_train_end(self):
        self.trainer.save_checkpoint(join(self.cfg.out_dir, "model.ckpt"))
        torch.save(self, join(self.cfg.out_dir, "model.pkl"))

    def configure_optimizers(self):
        optimizer = hydra.utils.instantiate(self.cfg.optimizer, filter(lambda p: p.requires_grad, self.parameters()), lr=self.learning_rate)
        return optimizer

    ################################            protected methods, can be overwritten by child
    def _setencoders(self):
        self.encoders = torch.nn.ModuleList(
            [
                hydra.utils.instantiate(
                    eval(f"self.cfg.encoder.enc{i}"),
                    input_dim=d,
                    z_dim=self.z_dim,
                    _recursive_=False,
                    _convert_ = "all"
                )
                for i, d in enumerate(self.input_dim)
            ]
        )

    def _setdecoders(self):
        self.decoders = torch.nn.ModuleList(
            [
                hydra.utils.instantiate(
                    eval(f"self.cfg.decoder.dec{i}"),
                    input_dim=d,
                    z_dim=self.z_dim,
                    _recursive_=False,
                    _convert_ = "all"
                )
                for i, d in enumerate(self.input_dim)
            ]
        )

    def _setprior(self):
        if self.model_name not in VARIATIONAL_MODELS or \
            (self.model_name in VARIATIONAL_MODELS and not self.sparse):

            self.prior = hydra.utils.instantiate(self.cfg.prior)

    def _unpack_batch(self, batch): # dataset returned other vars than x, need to unpack
        if isinstance(batch[0], list): 
            batch_x, batch_y, *other = batch
        else: 
            batch_x, batch_y, other = batch, None, None
        return batch_x, batch_y, other 

    def _set_batch_labels(self, labels): # for passing labels to encoder/decoder
        for i in range(len(self.encoders)):
            if isinstance(self.encoders[i], ConditionalVariationalEncoder):
                self.encoders[i].set_labels(labels)

        for i in range(len(self.decoders)):
            if isinstance(self.decoders[i], ConditionalVariationalDecoder):
                self.decoders[i].set_labels(labels)

        if hasattr(self, "private") and hasattr(self, "private_encoders"): 
            if self.private: 
                for i in range(len(self.private_encoders)):
                    if isinstance(self.private_encoders[i], ConditionalVariationalEncoder):
                        self.private_encoders[i].set_labels(labels)
        if hasattr(self, "conditional"): 
            if self.conditional: 
                self.set_labels(labels)

    ################################            private methods
    def __initcfg(self, old_cfg, new_cfg, at_fit=False, is_print=False):        
        updated_cfg = self.__updateconfig(old_cfg, new_cfg)

        if not at_fit and self.z_dim is not None:   # overrides hydra config... passed arg has precedence
            updated_cfg.model.z_dim = self.z_dim

        self.cfg = self.__checkconfig(updated_cfg)

        self.__dict__.update(self.cfg.model)

        if is_print:
            print("MODEL: ", self.model_name)
            self.print_config() #TODO: put this in debug mode logging

        if all(k in self.cfg.model for k in ["seed_everything", "seed"]):
            pl.seed_everything(self.cfg.model.seed, workers=True)

        if not at_fit or ("encoder" in new_cfg.keys()):
            self._setencoders()

        if not at_fit or ("decoder" in new_cfg.keys()):
            self._setdecoders()

        if not at_fit or ("prior" in new_cfg.keys()):
            self._setprior()
        
        if at_fit:
            self.create_folder(self.cfg.out_dir)
            self.save_config()

    def __updateconfig(self, orig, update):
        OmegaConf.set_struct(orig, True)
        with open_dict(orig):
            # update default cfg with user config
            if update is not None:
                update_keys = list(set(update.keys()) & set(CONFIG_KEYS))
                orig = update_dict(orig, update, l=update_keys)

            # update encoder/decoder config
            for i, d in enumerate(self.input_dim):
                enc_key = f"enc{i}"
                if enc_key not in orig.encoder.keys():
                    if update is not None and "encoder" in update.keys() and \
                        enc_key in update.encoder.keys(): # use user-defined
                        orig.encoder[enc_key] = update.encoder[enc_key].copy()
                    else: # use default
                        orig.encoder[enc_key] = orig.encoder.default.copy()

                dec_key = f"dec{i}"
                if dec_key not in orig.decoder.keys():
                    if update is not None and "decoder" in update.keys() and \
                        dec_key in update.decoder.keys(): # use user-defined
                        orig.decoder[dec_key] = update.decoder[dec_key].copy()
                    else: # use default
                        orig.decoder[dec_key] = orig.decoder.default.copy()
        if update is not None and update.get('out_dir'):
            orig.out_dir = update.out_dir
        return orig

    def __checkconfig(self, cfg):

        cfg_dict = OmegaConf.to_container(cfg)

        try:
            cfg_dict = config_schema.validate(cfg_dict)
        except SchemaError as se:
            raise ConfigError(se) from None
            
        if self.model_name == MODEL_JMVAE and len(self.input_dim) != 2:
            raise InputError('JMVAE expects two len(input_dim) == 2')

        if cfg.datamodule.dataset._target_ not in SUPPORTED_DATASETS:
            print("WARNING: Training on user specified dataset, please ensure that the dataset is compatible with the multi-view-AE framework")
            self.is_index_ds = cfg.datamodule.dataset.is_path_ds
        else:
            pattern = re.compile(r'multiviewae\.base\.datasets\.IndexMVDataset')
            if bool(pattern.match(cfg.datamodule.dataset._target_)):
                self.is_index_ds = True
            else:
                self.is_index_ds = False

        pattern = re.compile(r'..*\.Decoder')
        for k in cfg.decoder.keys():
            if eval(f"cfg.decoder.{k}.dec_dist._target_") in \
            ["multiviewae.base.distributions.Default", \
            "multiviewae.base.distributions.Bernoulli"]:
                if not bool(pattern.match(eval(f"cfg.decoder.{k}._target_"))):
                    raise ConfigError(f"{k}: must use non-variational Decoder if decoder dist is Default/Bernoulli.")

        if self.model_name in [MODEL_AE] + ADVERSARIAL_MODELS:
            pattern1 = re.compile(r'..*\.*VariationalEncoder')
            pattern2 = re.compile(r'multiviewae\.base\.distributions\..*Normal')
            for k in cfg.encoder.keys():
                if bool(pattern1.match(eval(f"cfg.encoder.{k}._target_"))):
                    raise ConfigError(f"{k}: must use non-variational encoder for adversarial models.")

                if bool(pattern2.match(eval(f"cfg.encoder.{k}.enc_dist._target_"))):
                    raise ConfigError(f"{k}: must not use Normal/MultivariateNormal encoder dist for adversarial models")

            pattern3 = re.compile(r'..*\.*VariationalDecoder')
            for k in cfg.decoder.keys():
                if bool(pattern3.match(eval(f"cfg.decoder.{k}._target_"))):
                    raise ConfigError(f"{k}: must use non-variational decoder for adversarial models.")

        if self.model_name in VARIATIONAL_MODELS:
            self.is_variational = True

            pattern = re.compile(r'..*\.*VariationalEncoder')
            pattern1 = re.compile(r'..*\.ConditionalVariational*')
            for k in cfg.encoder.keys():
                if not bool(pattern.match(eval(f"cfg.encoder.{k}._target_"))):
                    raise ConfigError(f"{k}: must use variational encoder for variational models")

                if cfg.prior._target_ != eval(f"cfg.encoder.{k}.enc_dist._target_"):
                    raise ConfigError('Encoder and prior must have the same distribution for variational models')
                
                if bool(pattern1.match(eval(f"cfg.encoder.{k}._target_"))) and "num_cat" not in eval(f"cfg.encoder.{k}"):
                    raise ConfigError('Condtional Variational Encoder must have the num_cat attribute')

            for k in cfg.decoder.keys():
                if bool(pattern1.match(eval(f"cfg.decoder.{k}._target_"))) and "num_cat" not in eval(f"cfg.decoder.{k}"):
                    raise ConfigError('Condtional Variational Decoder must have the num_cat attribute')

        if cfg.prior._target_ == "multiviewae.base.distributions.Normal":
            if not isinstance(cfg.prior.loc, (int, float)):
                raise ConfigError("loc must be int/float for Normal prior dist")

            if not isinstance(cfg.prior.scale, (int, float)):
                raise ConfigError("scale must be int/float for Normal prior dist")

        else:   # MultivariateNormal
            if isinstance(cfg.prior.loc, (int, float)):
                cfg.prior.loc = [cfg.prior.loc] * cfg.model.z_dim

            if isinstance(cfg.prior.scale, (int, float)):
                cfg.prior.scale = [cfg.prior.scale] * cfg.model.z_dim

            if  len(cfg.prior.loc) != len(cfg.prior.scale):
                raise ConfigError("loc and scale must have the same length for MultivariateNormal prior dist")

            if len(cfg.prior.loc) != cfg.model.z_dim:
                raise ConfigError("loc and scale must have the same length as z_dim for MultivariateNormal prior dist")

        # should be always false for non-sparse models
        if self.model_name not in SPARSE_MODELS:
            cfg.model.sparse = False    # TODO: log warning if changing overriding value

        return cfg

    def __step(self, batch, batch_idx, stage):
        batch_x, batch_y, other = self._unpack_batch(batch)
        self._set_batch_labels(batch_y)
                
        fwd_return = self.forward(batch_x)
        loss = self.loss_function(batch_x, fwd_return)
        for loss_n, loss_val in loss.items():
            self.log(
                f"{stage}_{loss_n}", loss_val, on_epoch=True, prog_bar=True, logger=True
            )
        return loss["loss"]

    def __predict(self, *data, input_modalities=None, output_modalities=None, labels=None, batch_size=None, is_recon=False):
        if any([isinstance(enc, ConditionalVariationalEncoder) for enc in self.encoders]) and labels is None: 
            raise InputError("no labels given for Conditional VAE")

        if any([isinstance(dec, ConditionalVariationalDecoder) for dec in self.decoders]) and labels is None: 
            raise InputError("no labels given for Conditional VAE")
        
        self._training = False
        for i in range(len(self.encoders)):
            self.encoders[i].training = False
        for i in range(len(self.decoders)):
            self.decoders[i].training = False

        data = list(data)
        if not self.is_index_ds:
            if not (len(data) == self.n_views):
                raise InputError("number of modalities must be equal to number of views")

            for i in range(self.n_views):
                data_dim = data[i].shape[1:]
                if len(data_dim) == 1:
                    data_dim = data_dim[0]
                if not (data_dim == self.input_dim[i]):
                    raise InputError("modality's shape must be equal to corresponding input_dim's shape")

        dataset = hydra.utils.instantiate(self.cfg.datamodule.dataset, data=data, labels=labels, n_views=self.n_views)
        if batch_size is None:
            if self.is_index_ds:
                batch_size = len(data[0])
            else:
                batch_size = data[0].shape[0]

        generator = DataLoader(dataset, batch_size=batch_size, shuffle=False)

        with torch.no_grad():
            z_ = None
            for batch_idx, local_batch in enumerate(generator):
                local_batchx, local_batchy, _ = self._unpack_batch(local_batch)
                self._set_batch_labels(local_batchy)

                local_batchx = [
                    local_batchx_.to(self.device) for local_batchx_ in local_batchx
                ]
                if input_modalities is None:
                    z = self.encode(local_batchx)
                else:
                    z = self.encode_subset(local_batchx, input_modalities)
                if self.sparse:
                    z = self.apply_threshold(z)
                if is_recon:
                    if output_modalities is None:
                        z = self.decode(z)
                    else:
                        z = self.decode_subset(z, output_modalities)

                z = [
                        [ d__._sample().cpu().detach().numpy() for d__ in d_ ]
                        if isinstance(d_, (list))
                        else
                        (d_.cpu().detach().numpy() if isinstance(d_, torch.Tensor)
                        else d_._sample().cpu().detach().numpy())
                        for d_ in z
                    ]

                if z_ is not None:
                    z_ = [
                            [ np.append(d_, d, axis=0) for d_, d in zip(p_,p) ]
                            if isinstance(p_, list) else np.append(p_, p, axis=0)
                            for p_, p in zip(z_, z)
                         ]
                else:
                    z_ = z
        return z_

    def predict_nll(self, *data, labels=None, batch_size=None):
        if not hasattr(self, "calc_nll"):
            raise NotImplementedError("predict_nll not implemented for this model")
        if any([isinstance(enc, ConditionalVariationalEncoder) for enc in self.encoders]) and labels is None: 
            raise InputError("no labels given for Conditional VAE")

        if any([isinstance(dec, ConditionalVariationalDecoder) for dec in self.decoders]) and labels is None: 
            raise InputError("no labels given for Conditional VAE")
        
        self._training = False
        for i in range(len(self.encoders)):
            self.encoders[i].training = False
        for i in range(len(self.decoders)):
            self.decoders[i].training = False

        data = list(data)
        if not self.is_index_ds:
            if not (len(data) == self.n_views):
                raise InputError("number of modalities must be equal to number of views")

            for i in range(self.n_views):
                data_dim = data[i].shape[1:]
                if len(data_dim) == 1:
                    data_dim = data_dim[0]
                if not (data_dim == self.input_dim[i]):
                    raise InputError("modality's shape must be equal to corresponding input_dim's shape")

        dataset = hydra.utils.instantiate(self.cfg.datamodule.dataset, data=data, labels=labels, n_views=self.n_views)
        if batch_size is None:
            if self.is_index_ds:
                batch_size = len(data[0])
            else:
                batch_size = data[0].shape[0]

        generator = DataLoader(dataset, batch_size=batch_size, shuffle=False)

        with torch.no_grad():
            ll = 0
            for batch_idx, local_batch in enumerate(generator):
                local_batchx, local_batchy, _ = self._unpack_batch(local_batch)
                self._set_batch_labels(local_batchy)

                local_batchx = [
                    local_batchx_.to(self.device) for local_batchx_ in local_batchx
                ]
                ll += self.calc_nll(local_batchx)
        return ll/(batch_idx+1)
################################################################################
class BaseModelVAE(BaseModelAE):
    """Base class for variational autoencoder models. Inherits from BaseModelAE.
    Args:
        model_name (str): Type of autoencoder model.
        cfg (str): Path to configuration file.
        input_dim (list): Dimensionality of the input data.
        z_dim (int): Number of latent dimensions.
    """
    @abstractmethod
    def __init__(
        self,
        model_name = None,
        cfg = None,
        input_dim = None,
        z_dim = None
    ):

        super().__init__(model_name=model_name,
                cfg=cfg,
                input_dim=input_dim,
                z_dim=z_dim)
        
    ################################            class methods
    def apply_threshold(self, z):
        """
        Implementation from: https://github.com/ggbioing/mcvae
        """

        keep = (self.__dropout() < self.threshold).squeeze().cpu()
        z_keep = []

        for _ in z:
            _ = _._sample()
            _[:, ~keep] = 0
            d = hydra.utils.instantiate(   
                self.cfg.encoder.default.enc_dist, loc=_, scale=1
            )
            z_keep.append(d)
            del _

        return z_keep

    ################################            protected methods
    def _setencoders(self):

        if self.sparse and self.threshold != 0.:
            self.log_alpha = torch.nn.Parameter(
                torch.FloatTensor(1, self.z_dim).normal_(0, 0.01)
            )
        else:
            self.sparse = False
            self.log_alpha = None
        
        #TODO: fix this annoying workaround
        if not hasattr(self, "u_dim"):
            self.u_dim = None
        if not hasattr(self, "w_dim"):
            self.w_dim = None
        if not hasattr(self, "multiple_latents"):
            self.multiple_latents = False

        self.encoders = torch.nn.ModuleList(
            [
                hydra.utils.instantiate(
                    eval(f"self.cfg.encoder.enc{i}"),
                    input_dim=d,
                    z_dim=self.z_dim,
                    sparse=self.sparse,
                    log_alpha=self.log_alpha,
                    u_dim=self.u_dim,
                    w_dim=self.w_dim,
                    multiple_latents=self.multiple_latents,
                    _recursive_=False,
                    _convert_="all"
                )
                for i, d in enumerate(self.input_dim)
            ]
        )

    ################################            private methods
    def __dropout(self):
        """
        Implementation from: https://github.com/ggbioing/mcvae
        """
        alpha = torch.exp(self.log_alpha.detach())
        return alpha / (alpha + 1)

################################################################################
class BaseModelAAE(BaseModelAE):
    """Base class for adversarial autoencoder models. Inherits from BaseModelAE.
    Args:
        model_name (str): Type of autoencoder model.
        cfg (str): Path to configuration file.
        input_dim (list): Dimensionality of the input data.
        z_dim (int): Number of latent dimensions.
    """
    is_wasserstein = False

    @abstractmethod
    def __init__(
        self,
        model_name = None,
        cfg = None,
        input_dim = None,
        z_dim = None
    ):

        super().__init__(model_name=model_name,
                        cfg=cfg,
                        input_dim=input_dim,
                        z_dim=z_dim)

        self.automatic_optimization = False

        self.discriminator = hydra.utils.instantiate(
            self.cfg.discriminator,
            input_dim=self.z_dim,
            output_dim=1,
            is_wasserstein=self.is_wasserstein,
            _convert_="all"
        )

    ################################            abstract methods
    @abstractmethod
    def forward_recon(self, x):
        raise NotImplementedError()

    @abstractmethod
    def forward_discrim(self, x):
        raise NotImplementedError()

    @abstractmethod
    def forward_gen(self, x):
        raise NotImplementedError()

    @abstractmethod
    def recon_loss(self, x, fwd_rtn):
        raise NotImplementedError()

    @abstractmethod
    def generator_loss(self, x, fwd_rtn):
        raise NotImplementedError()

    @abstractmethod
    def discriminator_loss(self, x, fwd_rtn):
        raise NotImplementedError()

    ################################        unused abstract methods
    def loss_function(self, x, fwd_rtn):
        pass

    def forward(self, x):
        pass

    ################################            LightningModule methods
    def training_step(self, batch, batch_idx):
        batch_x, batch_y, other = self._unpack_batch(batch)
        self._set_batch_labels(batch_y)

        loss = self.__optimise_batch(batch_x)
        for loss_n, loss_val in loss.items():
            self.log(
                f"train_{loss_n}", loss_val, on_epoch=True, prog_bar=True, logger=True
            )
        return loss["loss"]

    def validation_step(self, batch, batch_idx):
        batch_x, batch_y, other = self._unpack_batch(batch)
        self._set_batch_labels(batch_y)

        loss = self.__validate_batch(batch_x)
        for loss_n, loss_val in loss.items():
            self.log(
                f"val_{loss_n}", loss_val, on_epoch=True, prog_bar=True, logger=True
            )
        return loss["loss"]

    def configure_optimizers(self):
        optimizers = []
        #Encoder optimizers
        [
            optimizers.append(
                hydra.utils.instantiate(self.cfg.optimizer,
                    list(self.encoders[i].parameters()), lr=self.learning_rate
                )
            )
            for i in range(self.n_views)
        ]
        #Decoder optimizers
        [
            optimizers.append(
                hydra.utils.instantiate(self.cfg.optimizer,
                    list(self.decoders[i].parameters()), lr=self.learning_rate
                )
            )
            for i in range(self.n_views)
        ]
        #Generator optimizers
        [
            optimizers.append(
                hydra.utils.instantiate(self.cfg.optimizer,
                    list(self.encoders[i].parameters()), lr=self.learning_rate
                )
            )
            for i in range(self.n_views)
        ]
        #Discriminator optimizers
        optimizers.append(
            hydra.utils.instantiate(self.cfg.optimizer,
                list(self.discriminator.parameters()), lr=self.learning_rate
            )
        )
        return optimizers

    ################################            private methods
    def __optimise_batch(self, local_batch):
        fwd_return = self.forward_recon(local_batch)
        loss_recon = self.recon_loss(local_batch, fwd_return)
        opts = self.optimizers()
        enc_opt = [opts.pop(0) for idx in range(self.n_views)]
        dec_opt = [opts.pop(0) for idx in range(self.n_views)]
        gen_opt = [opts.pop(0) for idx in range(self.n_views)]
        disc_opt = opts[0]
        [optimizer.zero_grad() for optimizer in enc_opt]
        [optimizer.zero_grad() for optimizer in dec_opt]
        self.manual_backward(loss_recon)
        [optimizer.step() for optimizer in enc_opt]
        [optimizer.step() for optimizer in dec_opt]

        fwd_return = self.forward_discrim(local_batch)
        loss_disc = self.discriminator_loss(fwd_return)
        disc_opt.zero_grad()
        self.manual_backward(loss_disc)
        disc_opt.step()
        if self.is_wasserstein:
            for p in self.discriminator.parameters():
                p.data.clamp_(-0.01, 0.01)
        fwd_return = self.forward_gen(local_batch)
        loss_gen = self.generator_loss(fwd_return)
        [optimizer.zero_grad() for optimizer in gen_opt]
        self.manual_backward(loss_gen)
        [optimizer.step() for optimizer in gen_opt]
        loss_total = loss_recon + loss_disc + loss_gen
        loss = {
            "loss": loss_total,
            "recon": loss_recon,
            "disc": loss_disc,
            "gen": loss_gen,
        }
        return loss

    def __validate_batch(self, local_batch):
        with torch.no_grad():
            self.eval()
            fwd_return = self.forward_recon(local_batch)
            loss_recon = self.recon_loss(local_batch, fwd_return)
            fwd_return = self.forward_discrim(local_batch)
            loss_disc = self.discriminator_loss(fwd_return)
            fwd_return = self.forward_gen(local_batch)
            loss_gen = self.generator_loss(fwd_return)
            loss_total = loss_recon + loss_disc + loss_gen
            loss = {
                "loss": loss_total,
                "recon": loss_recon,
                "disc": loss_disc,
                "gen": loss_gen,
            }
        return loss





# constants file
MODEL_AE        = "AE"
MODEL_MAAE       = "mAAE"
MODEL_MWAE      = "mWAE"
MODEL_MCVAE     = "mcVAE"
MODEL_MVAE      = "mVAE"
MODEL_JMVAE     = "JMVAE"
MODEL_MEMVAE    = "me_mVAE"
MODEL_MMVAE     = "mmVAE"
MODEL_MVTCAE    = "mvtCAE"
MODEL_DVCCA     = "DVCCA"
MODEL_MOPOEVAE  = "MoPoEVAE"
MODEL_MMJSD     = "mmJSD"
MODEL_WEIGHTEDMVAE = "weighted_mVAE"
MODEL_DMVAE = "DMVAE"
MODEL_WEIGHTEDDMVAE = "weighted_DMVAE"
MODEL_MMVAEPLUS = "mmVAEPlus"
MODEL_DCCAE = "DCCAE"
MODELS = [
            MODEL_AE,
            MODEL_MAAE,
            MODEL_MWAE,
            MODEL_MCVAE,
            MODEL_JMVAE,
            MODEL_MVAE,
            MODEL_MEMVAE,
            MODEL_MMVAE,
            MODEL_MVTCAE,
            MODEL_DVCCA,
            MODEL_MOPOEVAE,
            MODEL_MMJSD,
            MODEL_WEIGHTEDMVAE,
            MODEL_DMVAE, 
            MODEL_WEIGHTEDDMVAE, 
            MODEL_MMVAEPLUS,
            MODEL_DCCAE,
        ]

ADVERSARIAL_MODELS = [
            MODEL_MAAE,
            MODEL_MWAE,
        ]

VARIATIONAL_MODELS = [
            MODEL_MCVAE,
            MODEL_MVAE,
            MODEL_JMVAE,
            MODEL_MEMVAE,
            MODEL_MMVAE,
            MODEL_MVTCAE,
            MODEL_DVCCA,
            MODEL_MOPOEVAE,
            MODEL_MMJSD,
            MODEL_WEIGHTEDMVAE,
            MODEL_DMVAE,
            MODEL_WEIGHTEDDMVAE,
            MODEL_MMVAEPLUS
        ]

SPARSE_MODELS = [
            MODEL_MCVAE,
            MODEL_MVAE,
            MODEL_DVCCA
        ]

CONFIG_KEYS = [
            "model",
            "datamodule",
            "encoder",
            "decoder",
            "prior",
            "trainer",
            "callbacks",
            "logger"
        ]

EPS = 1e-8


import random
import numpy as np
import pytorch_lightning as pl
import hydra
from torch.utils.data import DataLoader

class MultiviewDataModule(pl.LightningDataModule):
    """LightningDataModule for multi-view data.

    Args:
        n_views (int): Number of views in the data.
        batch_size (int): Batch size.
        is_validate (bool): Whether to use a validation set.
        train_size (float): Proportion of batch to use for training between 0 and 1. Remainder of batch is used for validation.
        data (list): Input data. list of torch.Tensors.
        labels (np.array): Dataset labels. 
    """
    def __init__(
            self,
            n_views,
            batch_size,
            is_validate,
            train_size,
            dataset,
            data,
            labels
        ):

        super().__init__()
        self.n_views = n_views
        self.batch_size = batch_size
        self.is_validate = is_validate
        self.train_size = train_size
        self.data = data  
        self.labels = labels 
        self.dataset = dataset
        if not isinstance(self.batch_size, int):
            self.batch_size = self.data[0].shape[0]

    def setup(self, stage):
        if self.is_validate:
            train_data, test_data, train_labels, test_labels = self.train_test_split()
            self.train_dataset = hydra.utils.instantiate(self.dataset, data=train_data, labels=train_labels, n_views=self.n_views)
            self.test_dataset = hydra.utils.instantiate(self.dataset, data=test_data, labels=test_labels, n_views=self.n_views)
        else:
            self.train_dataset = hydra.utils.instantiate(self.dataset, data=self.data, labels=self.labels, n_views=self.n_views) 
            self.test_dataset = None
        del self.data

    def train_test_split(self):

        N = self.data[0].shape[0]
        train_idx = list(random.sample(range(N), int(N * self.train_size)))
        test_idx = np.setdiff1d(list(range(N)), train_idx)

        train_data = []
        test_data = []
        for dt in self.data:
            train_data.append(dt[train_idx, :])
            test_data.append(dt[test_idx, :])

        train_labels = None
        test_labels = None
        if self.labels is not None:
            train_labels = self.labels[train_idx]
            test_labels = self.labels[test_idx]

        return train_data, test_data, train_labels, test_labels

    def train_dataloader(self):
        return DataLoader(self.train_dataset, batch_size=self.batch_size, shuffle=False, num_workers=0) #use default num_workers for now, problem in windows! 

    def val_dataloader(self):
        if self.is_validate:
            return DataLoader(
                self.test_dataset, batch_size=self.batch_size, shuffle=False, num_workers=0)
        return None


class IndexDataModule(MultiviewDataModule):
    """LightningDataModule for multi-view data.

    Args:
        n_views (int): Number of views in the data.
        batch_size (int): Batch size.
        is_validate (bool): Whether to use a validation set.
        train_size (float): Proportion of batch to use for training between 0 and 1. Remainder of batch is used for validation.
        data (list): Input data. list of identifiers to load data from.
        labels (np.array): Dataset labels. 
    """

    def __init__(
            self,
            n_views, 
            batch_size,
            is_validate,
            train_size,
            dataset,
            data,
            labels,
        ):
        data_ = data[0]
        if not isinstance(batch_size, int):
            batch_size = len(data_)

        super().__init__(n_views=n_views, batch_size=batch_size, is_validate=is_validate, train_size=train_size, 
                         dataset=dataset, data=data_, labels=labels)

    def train_test_split(self):

        N = len(self.data)
        train_idx = list(random.sample(range(N), int(N * self.train_size)))
        test_idx = list(set(list(range(N))) -  set(train_idx))
        data = self.data

        train_data = [data[i] for i in train_idx]
        test_data = [data[i] for i in test_idx]

        train_labels = None
        test_labels = None
        if self.labels is not None:
            train_labels = self.labels[train_idx]
            test_labels = self.labels[test_idx]

        return [train_data], [test_data], train_labels, test_labels



import torch
import hydra

from ..base.constants import MODEL_WEIGHTEDMVAE
from ..base.base_model import BaseModelVAE
from ..base.representations import weightedProductOfExperts

class weighted_mVAE(BaseModelVAE):
    r"""
    Generalised Product-of-Experts Variational Autoencoder (gPoE-MVAE).

    Args:
        cfg (str): Path to configuration file. Model specific parameters in addition to default parameters:
            
            - model.beta (int, float): KL divergence weighting term.
            - model.private (bool): Whether to include private view-specific latent dimensions.
            - model.sparse (bool): Whether to enforce sparsity of the encoding distribution.
            - model.threshold (float): Dropout threshold applied to the latent dimensions. Default is 0.
            - encoder.default._target_ (multiviewae.architectures.mlp.VariationalEncoder): Type of encoder class to use.
            - encoder.default.enc_dist._target_ (multiviewae.base.distributions.Normal, multiviewae.base.distributions.MultivariateNormal): Encoding distribution.
            - decoder.default._target_ (multiviewae.architectures.mlp.VariationalDecoder): Type of decoder class to use.
            - decoder.default.init_logvar (int, float): Initial value for log variance of decoder.
            - decoder.default.dec_dist._target_ (multiviewae.base.distributions.Normal, multiviewae.base.distributions.MultivariateNormal): Decoding distribution.
        
        input_dim (list): Dimensionality of the input data.
        z_dim (int): Number of latent dimensions.

    References
    ----------
    Cao, Y., & Fleet, D. (2014). Generalized Product of Experts for Automatic and Principled Fusion of Gaussian Process Predictions. arXiv.
    Lawry Aguila, A., Chapman, J., Altmann, A. (2023). Multi-modal Variational Autoencoders for normative modelling across multiple imaging modalities. arXiv

    """

    def __init__(
        self,
        cfg = None,
        input_dim = None,
        z_dim = None
    ):
        super().__init__(model_name=MODEL_WEIGHTEDMVAE,
                        cfg=cfg,
                        input_dim=input_dim,
                        z_dim=z_dim)

        self.join_z = weightedProductOfExperts()
        #check if self.private is attribute, if not set to False
        if not hasattr(self, 'private'):
            self.private = False
        if self.private:
            tmp_weight = torch.FloatTensor(len(input_dim), self.z_dim - self.s_dim).fill_(1/len(input_dim))
            self.poe_weight = torch.nn.Parameter(data=tmp_weight, requires_grad=True)
        else:
            tmp_weight = torch.FloatTensor(len(input_dim), self.z_dim).fill_(1/len(input_dim))
            self.poe_weight = torch.nn.Parameter(data=tmp_weight, requires_grad=True)

    def encode(self, x):
        r"""Forward pass through encoder networks.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            Returns a combination of the following depending on the training stage and model type: 
            qz_x (list):  Single element list containing the PoE encoding distribution for self.private=False.

            qc_x (list): Single element list containing the PoE shared encoding distribution.
            qs_xs (list): list of encoding distributions for private latents.
            qscs_xs (list): list containing combined shared and private latents.
            qcs_xs (list): list containing encoding distributions for shared latent dimensions for each view.
        """
        if not hasattr(self, 'private'):
            self.private = False
        if self.private:
            qs_xs = []
            qcs_xs = []
            mu_s = []
            logvar_s = []
            mu_c = []
            logvar_c = []
            for i in range(self.n_views):
                mu, logvar = self.encoders[i](x[i])
                mu_s.append(mu[:,:self.s_dim])
                logvar_s.append(logvar[:,:self.s_dim])
                mu_c.append(mu[:,self.s_dim:])
                logvar_c.append(logvar[:,self.s_dim:])

                qs_x = hydra.utils.instantiate(
                eval(f"self.cfg.encoder.enc{i}.enc_dist"), loc=mu[:,:self.s_dim], logvar=logvar[:,:self.s_dim]
                )
                qs_xs.append(qs_x)
                qc_x = hydra.utils.instantiate(
                eval(f"self.cfg.encoder.enc{i}.enc_dist"), loc=mu[:,self.s_dim:], logvar=logvar[:,self.s_dim:]
                )
                qcs_xs.append(qc_x)

            mu_c = torch.stack(mu_c)
            logvar_c = torch.stack(logvar_c)
       
            mu_c, logvar_c = self.join_z(mu_c, logvar_c, self.poe_weight)
            qc_x = hydra.utils.instantiate(
                self.cfg.encoder.default.enc_dist, loc=mu_c, logvar=logvar_c
            )
            with torch.no_grad():
                self.poe_weight = self.poe_weight.clamp_(0, +1)
            qscs_xs = []
            for i in range(self.n_views):       
                mu_sc = torch.cat((mu_s[i], mu_c), 1)
                logvar_sc = torch.cat((logvar_s[i], logvar_c), 1)
                qsc_x = hydra.utils.instantiate( 
                eval(f"self.cfg.encoder.enc{i}.enc_dist"), loc=mu_sc, logvar=logvar_sc
                )
                qscs_xs.append(qsc_x)

            if self._training:
                return [[qc_x], qcs_xs, qs_xs, qscs_xs]
            return qscs_xs

        mu = []
        logvar = []
        for i in range(self.n_views):
            mu_, logvar_ = self.encoders[i](x[i])
            mu.append(mu_)
            logvar.append(logvar_)
        mu = torch.stack(mu)
        logvar = torch.stack(logvar)
        mu_out, logvar_out = self.join_z(mu, logvar, self.poe_weight)
        qz_x = hydra.utils.instantiate(
            self.cfg.encoder.default.enc_dist, loc=mu_out, logvar=logvar_out
        )
        with torch.no_grad():
            self.poe_weight = self.poe_weight.clamp_(0, +1)
        return [qz_x]

    def decode(self, qz_x):
        r"""Forward pass of joint latent dimensions through decoder networks.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            (list): A nested list of decoding distributions, px_zs. The outer list has a single element indicating the shared latent dimensions. 
            The inner list is a n_view element list with the position in the list indicating the decoder index.
        """  
        px_zs = []
        for i in range(self.n_views):
            if self.private:
                px_z = self.decoders[i](qz_x[i]._sample(training=self._training, return_mean=self.return_mean))
            else:
                px_z = self.decoders[i](qz_x[0]._sample(training=self._training, return_mean=self.return_mean))
            px_zs.append(px_z)
        return [px_zs]

    def forward(self, x):
        r"""Apply encode and decode methods to input data to generate the joint latent dimensions and data reconstructions. 
        
        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            fwd_rtn (dict): dictionary containing encoding and decoding distributions.
        """
        if self.private:
            qc_x, qcs_xs, qs_xs, qscs_xs = self.encode(x)
            px_zs = self.decode(qscs_xs)
            fwd_rtn = {"px_zs": px_zs, "qcs_xs": qcs_xs, "qs_xs": qs_xs, "qc_x": qc_x}
        else:
            qz_x = self.encode(x)
            px_zs = self.decode(qz_x)
            fwd_rtn = {"px_zs": px_zs, "qz_x": qz_x}
        return fwd_rtn

    def calc_kl(self, qz_x):
        r"""Calculate KL-divergence loss.

        Args:
            qz_x (list): Single element list containing joint encoding distribution.

        Returns:
            (torch.Tensor): KL-divergence loss.
        """
        kl = qz_x[0].kl_divergence(self.prior).mean(0).sum()
        return self.beta * kl
    
    def calc_kl_separate(self, qc_xs):
        r"""Calculate KL-divergence loss.

        Args:
            qc_xs (list): list of encoding distributions for private/shared latent dimensions for each view.

        Returns:
            (torch.Tensor): KL-divergence loss.
        """
        kl = 0
        for i in range(self.n_views):
            kl += qc_xs[i].kl_divergence(self.prior).mean(0).sum()
        return self.beta * kl/self.n_views
    
    def calc_ll(self, x, px_zs):
        r"""Calculate log-likelihood loss.

        Args:
            x (list): list of input data of type torch.Tensor.
            px_zs (list): list of decoding distributions.

        Returns:
            ll (torch.Tensor): Log-likelihood loss.
        """
        ll = 0
        for i in range(self.n_views):
            ll += px_zs[0][i].log_likelihood(x[i]).mean(0).sum() #first index is latent, second index is view
        return ll/self.n_views

    def loss_function(self, x, fwd_rtn):
        r"""Calculate Multimodal VAE loss.
        
        Args:
            x (list): list of input data of type torch.Tensor.
            fwd_rtn (dict): dictionary containing encoding and decoding distributions.

        Returns:
            losses (dict): dictionary containing each element of the MVAE loss.
        """
        px_zs = fwd_rtn["px_zs"]
        ll = self.calc_ll(x, px_zs)
        if self.private:
            qs_xs = fwd_rtn["qs_xs"]
            qc_x = fwd_rtn["qc_x"]
            kl = self.calc_kl_separate(qs_xs) #calc kl for private latents
            kl += self.calc_kl(qc_x) #calc kl for shared latents
            total = kl - ll
            losses = {"loss": total, "kl": kl, "ll": ll}
            return losses
        else:
            qz_x = fwd_rtn["qz_x"]
            kl = self.calc_kl(qz_x)
            total = kl - ll
            losses = {"loss": total, "kl": kl, "ll": ll}
            return losses




import torch
import hydra
import numpy as np
from ..base.constants import MODEL_JMVAE
from ..base.base_model import BaseModelVAE

class JMVAE(BaseModelVAE):
    r"""
    JMVAE-kl.

    Args:
        cfg (str): Path to configuration file. Model specific parameters in addition to default parameters:
            - alpha (float): Weighting of KL-divergence loss from individual encoders.
            - encoder.default._target_ (multiviewae.architectures.mlp.VariationalEncoder): Type of Encoder to use.
            - encoder.default.enc_dist._target_ (multiviewae.base.distributions.Normal, multiviewae.base.distributions.MultivariateNormal): Encoding distribution.
            - decoder.default._target_ (multiviewae.architectures.mlp.VariationalDecoder): Type of decoder class to use.
            - decoder.default.init_logvar(int, float): Initial value for log variance of decoder.
            - decoder.default.dec_dist._target_ (multiviewae.base.distributions.Normal, multiviewae.base.distributions.MultivariateNormal): Decoding distribution.
        
        input_dim (list): Dimensionality of the input data.
        z_dim (int): Number of latent dimensions.

    References
    ----------
    Suzuki, Masahiro & Nakayama, Kotaro & Matsuo, Yutaka. (2016). Joint Multimodal Learning with Deep Generative Models.
    """

    def __init__(
        self,
        cfg = None,
        input_dim = None,
        z_dim = None
    ):
        super().__init__(model_name=MODEL_JMVAE,
                        cfg=cfg,
                        input_dim=input_dim,
                        z_dim=z_dim)

    def _setencoders(self):
        r"""Set the joint and individual encoder networks.
        """
        self.encoders = torch.nn.ModuleList(
              [hydra.utils.instantiate( 
                    self.cfg.encoder.default,
                    input_dim=self.input_dim[0]+self.input_dim[1],
                    z_dim=self.z_dim,
                    sparse=False,
                    log_alpha=None,
                    _recursive_=False,
                    _convert_="all"
                )]
                +
            [
                hydra.utils.instantiate(
                    eval(f"self.cfg.encoder.enc{i}"),
                    input_dim=d,
                    z_dim=self.z_dim,
                    sparse=False,
                    log_alpha=None,
                    _recursive_=False,
                    _convert_="all"
                )
                for i, d in enumerate(self.input_dim)
            ]
        )

    def encode(self, x):
        r"""Forward pass through joint encoder network. 

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            (list): Single element list containing joint encoding distribution, qz_xy.
        """
        mu, logvar = self.encoders[0](torch.cat((x[0], x[1]),dim=1)) #TODO: current implementation only works for 2D data for both modalities
        qz_xy = hydra.utils.instantiate(  
            self.cfg.encoder.default.enc_dist, loc=mu, logvar=logvar
        )
        return [qz_xy]

    def encode_separate(self, x):
        r"""Forward pass through separate encoder networks. 

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            qz_x: Encoding distribution for modality X.
            qz_y: Encoding distribution for modality Y.
        """
        mu, logvar = self.encoders[1](x[0])
        qz_x = hydra.utils.instantiate(    
            self.cfg.encoder.enc0.enc_dist, loc=mu, logvar=logvar
        )
        mu, logvar = self.encoders[2](x[1])
        qz_y = hydra.utils.instantiate(   
            self.cfg.encoder.enc1.enc_dist, loc=mu, logvar=logvar
        )
        return qz_x, qz_y

    def decode(self, qz_x):
        r"""Forward pass of joint latent dimensions through decoder networks.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            (list): A nested list of decoding distributions, px_zs. The outer list has a single element indicating the shared latent dimensions. 
            The inner list is a 2 element list with the position in the list indicating the decoder index.
        """       
        px_zs = []
        for i in range(self.n_views):
            px_z = self.decoders[i](qz_x[0]._sample(training=self._training, return_mean=self.return_mean))
            px_zs.append(px_z)
        return [px_zs]

    def forward(self, x):
        r"""Apply encode and decode methods to input data to generate latent dimensions and data reconstructions. 
        
        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            fwd_rtn (dict): dictionary containing encoding and decoding distributions.
        """
        qz_xy = self.encode(x)
        qz_x, qz_y = self.encode_separate(x)
        px_z, py_z = self.decode(qz_xy)[0]
        fwd_rtn = {"px_z": px_z, "py_z": py_z, "qz_x": qz_x, "qz_y": qz_y, "qz_xy": qz_xy}
        return fwd_rtn

    def calc_kl(self, qz_xy, qz_x, qz_y):
        r"""Calculate JMVAE-kl KL-divergence loss.

        Args:
            qz_xy (list): Single element list containing shared encoding distribution.
            qz_x (list): Single element list containing encoding distribution for modality X.
            qz_y (list): Single element list containing encoding distribution for modality Y.


        Returns:
            (torch.Tensor): KL-divergence loss.
        """

        kl_prior = qz_xy[0].kl_divergence(self.prior).mean(0).sum()
        kl_qz_x = qz_xy[0].kl_divergence(qz_x).mean(0).sum()
        kl_qz_y = qz_xy[0].kl_divergence(qz_y).mean(0).sum()

        return kl_prior + self.alpha*(kl_qz_x + kl_qz_y)

    def calc_ll(self, x, px_zs):
        r"""Calculate log-likelihood loss.

        Args:
            x (list): list of input data of type torch.Tensor.
            px_zs (list): list of decoding distributions.

        Returns:
            ll (torch.Tensor): Log-likelihood loss.
        """
        ll = 0
        for i in range(self.n_views):
            ll += px_zs[i].log_likelihood(x[i]).mean(0).sum()
        return ll

    def loss_function(self, x, fwd_rtn):
        r"""Calculate JMVAE-kl loss.
        
        Args:
            x (list): list of input data of type torch.Tensor.
            fwd_rtn (dict): dictionary containing encoding and decoding distributions.

        Returns:
            losses (dict): dictionary containing each element of the JMVAE loss.
        """
        px_z = fwd_rtn["px_z"]
        py_z = fwd_rtn["py_z"]
        qz_x = fwd_rtn["qz_x"]
        qz_y = fwd_rtn["qz_y"]
        qz_xy = fwd_rtn["qz_xy"]

        kls = self.calc_kl(qz_xy, qz_x, qz_y)
        ll = self.calc_ll(x, [px_z, py_z])

        if self.current_epoch > self.warmup - 1:
            annealing_factor = 1
        else:
            annealing_factor = self.current_epoch/self.warmup

        total = annealing_factor*kls - ll

        losses = {"loss": total, "kls": kls, "ll": ll}
        return losses

    def calc_nll(self, x, K=1000, batch_size_K=100):
        r"""Calculate negative log-likelihood used to evaluate model performance.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            nll (torch.Tensor): Negative log-likelihood.
        """
        mu, logvar = self.encoders[0](torch.cat((x[0], x[1]),dim=1)) #TODO: current implementation only works for 2D data for both modalities
        qz_xy = hydra.utils.instantiate(  
            self.cfg.encoder.default.enc_dist, loc=mu, logvar=logvar
        )
        # And sample from the posterior
        zs = qz_xy.rsample([K])  # shape K x n_data x latent_dim
        ll = 0
        start_idx = 0
        stop_idx = min(start_idx + batch_size_K, K)
        lnpxs = []
        while start_idx < stop_idx:
            zs_ = zs[start_idx:stop_idx]
            lpx_zs = 0
            for j in range(self.n_views):
                px_z = self.decoders[j](zs_)
                lpx_zs += px_z.log_likelihood(x[j]).sum(dim=-1)

            lpz = self.prior.log_likelihood(zs_).sum(dim=-1)
            lqz_xy = qz_xy.log_likelihood(zs_).sum(dim=-1)
            ln_px = torch.logsumexp(lpx_zs + lpz - lqz_xy, dim=0)
            lnpxs.append(ln_px)
            start_idx += batch_size_K
            stop_idx = min(stop_idx + batch_size_K, K)
        lnpxs = torch.stack(lnpxs)
        ll += (torch.logsumexp(lnpxs, dim=0) - np.log(K)).mean()
        return -ll


import math
import torch
import hydra

from ..base.constants import MODEL_MMVAEPLUS
from ..base.base_model import BaseModelVAE
from ..base.distributions import Default
import numpy as np
#set numpy seed
np.random.seed(0)

#TODO: check if private priors are used at all here??

class mmVAEPlus(BaseModelVAE):
    r"""
    Mixture-of-Experts Multimodal Variational Autoencoder (MMVAE). 

    Code is based on: https://github.com/iffsid/mmvae

    Args:
        cfg (str): Path to configuration file. Model specific parameters in addition to default parameters:
            
            - model.K (int): Number of samples to take from encoding distribution.
            - model.DREG_loss (bool): Whether to use DReG estimator when using large K value. 
            - encoder.default._target_ (multiviewae.architectures.mlp.VariationalEncoder): Type of encoder class to use.
            - encoder.default.enc_dist._target_ (multiviewae.base.distributions.Normal, multiviewae.base.distributions.MultivariateNormal): Encoding distribution.
            - decoder.default._target_ (multiviewae.architectures.mlp.VariationalDecoder): Type of decoder class to use.
            - decoder.default.init_logvar (int, float): Initial value for log variance of decoder.
            - decoder.default.dec_dist._target_ (multiviewae.base.distributions.Normal, multiviewae.base.distributions.MultivariateNormal): Decoding distribution.
            
        input_dim (list): Dimensionality of the input data.
        z_dim (int): Number of latent dimensions. 
    
    References
    ----------
    Shi, Y., Siddharth, N., Paige, B., & Torr, P.H. (2019). Variational Mixture-of-Experts Autoencoders for Multi-Modal Deep Generative Models. ArXiv, abs/1911.03393.
    """

    def __init__(
        self,
        cfg = None,
        input_dim = None,
        z_dim = None
    ):
        super().__init__(model_name=MODEL_MMVAEPLUS,
                        cfg=cfg,
                        input_dim=input_dim,
                        z_dim=z_dim)
    
        #create the prior parameters
        self.mean_priors_private = []
        self.logvar_priors_private = []

        for i in range(self.n_views):
            #set parameters of private priors
            mean_prior = torch.nn.Parameter(torch.zeros(1, self.w_dim), requires_grad=False).to(self.device)
            logvar_prior = torch.nn.Parameter(torch.zeros(1, self.w_dim), requires_grad=self.learn_private_prior).to(self.device)
            self.mean_priors_private.append(mean_prior)
            self.logvar_priors_private.append(logvar_prior)
        #set parameters of shared priors
        mean_prior = torch.nn.Parameter(torch.zeros(1, self.u_dim+self.w_dim), requires_grad=False).to(self.device) 
        logvar_prior = torch.nn.Parameter(torch.zeros(1, self.u_dim+self.w_dim), requires_grad=self.learn_shared_prior).to(self.device) 
        self.mean_priors_shared = mean_prior
        self.logvar_priors_shared = logvar_prior

    def encode(self, x):
        r"""Forward pass through encoder networks.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            (list): list of encoding distributions for shared (qu_xs) and private (qw_xs) latent dimensions during training, otherwise return samples from encoding distributions.
        """
        qw_xs = []
        qu_xs = []
        for i in range(self.n_views):
            mu_u, logvar_u, mu_w, logvar_w = self.encoders[i](x[i])
            #private latent distribution
            qw_x = hydra.utils.instantiate(
                eval(f"self.cfg.encoder.enc{i}.enc_dist"), loc=mu_w, logvar=logvar_w
            )
            #shared latent distribution
            qu_x = hydra.utils.instantiate(
                eval(f"self.cfg.encoder.enc{i}.enc_dist"), loc=mu_u, logvar=logvar_u
            )

            qw_xs.append(qw_x)
            qu_xs.append(qu_x)
        if self._training:
            return [qw_xs, qu_xs]
        zss = []
        for i in range(self.n_views):

            u_x = qu_xs[i]._sample() #shared sample
            w_x = qw_xs[i]._sample() #private sample
            z_x = torch.cat((u_x, w_x), dim=-1)
            z_x = Default(x=z_x)

            zss.append(z_x)
        return zss
    
    def encode_subset(self, x, subset):
        r"""Forward pass through encoder networks for a subset of modalities. For modalities not in subset, shared latents are sampled from a random modality and 
        private latents from the shared prior.

        Args:
            x (list): list of input data of type torch.Tensor.
            subset (list): list of modalities to encode.

        Returns:
            (list): list of samples from encoding distributions.
        """    
        qw_xs = []
        qu_xs = []

        for i in range(self.n_views):
            if i in subset:
                mu_u, logvar_u, mu_w, logvar_w = self.encoders[i](x[i])
                #private latent distribution
                qw_x = hydra.utils.instantiate(
                    eval(f"self.cfg.encoder.enc{i}.enc_dist"), loc=mu_w, logvar=logvar_w
                )
                #shared latent distribution
                qu_x = hydra.utils.instantiate(
                    eval(f"self.cfg.encoder.enc{i}.enc_dist"), loc=mu_u, logvar=logvar_u
                )
            else: 
                # Choose one of the subset modalities at random
                mod = np.random.choice(subset)
                mu_u, logvar_u, mu_w, logvar_w = self.encoders[mod](x[mod])
                #shared latent distribution
                qu_x = hydra.utils.instantiate(
                    eval(f"self.cfg.encoder.enc{mod}.enc_dist"), loc=mu_u, logvar=logvar_u
                )
                #private latent distribution
                #sample private latent from shared prior
                qw_x = hydra.utils.instantiate(self.cfg.prior, loc=self.mean_priors_shared[:,self.u_dim:], logvar=self.logvar_priors_shared[:,self.u_dim:])
            qw_xs.append(qw_x)
            qu_xs.append(qu_x)
        if self._training:
            return [qw_xs, qu_xs]
        zss = []
        for i in range(self.n_views):
            w_x = qw_xs[i]._sample() #private sample
            u_x = qu_xs[i]._sample() #shared sample
            shape_ = list(w_x.shape)
            shape_[0] = u_x.shape[0]
            w_x = w_x.expand(shape_)
            z_x = torch.cat((u_x, w_x), dim=-1)
            z_x = Default(x=z_x)
            zss.append(z_x)
        return zss
    
    def decode(self, zss):
        r"""Forward pass through decoder networks. Each latent is passed through all of the decoders.

        Args:
            zss (list): list of latent samples if not training or list containing qw_xs (list of private encoding distributions) and
            qu_xs (list of shared encoding distributions) if training.

        Returns:
            (list): A nested list of decoding distributions. The outer list has a n_view element indicating latent dimensions index. 
            The inner list is a n_view element list with the position in the list indicating the decoder index.
        """    
        px_zs = []
        if self._training:
            qw_xs, qu_xs = zss[0], zss[1]
            for i in range(self.n_views):
                px_zs_inner = []
                for j in range(self.n_views): 
                    if i == j:
                        u_x = qu_xs[i].rsample(torch.Size([self.K])) #shared sample
                        w_x = qw_xs[j].rsample(torch.Size([self.K])) #private sample
                        z_x = torch.cat((u_x, w_x), dim=-1)
                    else:
                        u_x = qu_xs[i].rsample(torch.Size([self.K])) #shared sample
                        prior = hydra.utils.instantiate(
                            eval(f"self.cfg.encoder.enc{j}.enc_dist"), loc=self.mean_priors_private[j], logvar=self.logvar_priors_private[j]
                        )
                        w = prior.rsample(torch.Size([self.K])).to(self.device) #private sample
                        w = w.expand(w_x.shape)
                        z_x = torch.cat((u_x, w), dim=-1)
                    px_z = self.decoders[j](z_x)
                    px_zs_inner.append(px_z)
                px_zs.append(px_zs_inner)
        else:
            for i in range(self.n_views):
                px_zs_inner = []
                for j in range(self.n_views): 
                    z_x = zss[i]._sample(training=self._training, return_mean=self.return_mean)
                    if i != j:
                        u_x = z_x[:,:self.u_dim]
                        prior = hydra.utils.instantiate(self.cfg.prior, loc=self.mean_priors_shared[:,self.u_dim:], logvar=self.logvar_priors_shared[:,self.u_dim:])
                        w = prior.rsample().to(self.device) #private sample
                        shape_ = list(w.shape)
                        shape_[0] = u_x.shape[0]
                        w = w.expand(shape_)

                        z_x = torch.cat((u_x, w), dim=-1)
                    px_z = self.decoders[j](z_x) 
                    px_zs_inner.append(px_z)
                px_zs.append(px_zs_inner)            
        return px_zs

    def decode_subset(self, zss, subset):
        r"""Forward pass through decoder networks for a subset of modalities. Each latent is passed through its own decoder.

        Args:
            zss (list): list of latent samples for each modality.
            subset (list): list of modalities to decode.

        Returns:
            (list): A list of decoding distributions for each modality in subset.
        """    
        px_zs = []
        for i, zs in enumerate(zss):
            if i in subset:
                px_z = self.decoders[i](zs._sample(training=self._training, return_mean=self.return_mean))
                px_zs.append(px_z)
        return [px_zs]

    def forward(self, x):
        r"""Apply encode and decode methods to input data to generate latent dimensions and data reconstructions. 
        
        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            fwd_rtn (dict): dictionary containing encoding (qw_xs and qu_xs) and decoding (px_zs) distributions.
        """
        if self._training:
            qw_xs, qu_xs = self.encode(x)[0], self.encode(x)[1]
            px_zs = self.decode([qw_xs, qu_xs])
        else:
            zss = self.encode(x)
            px_zs = self.decode(zss)
        return {"qw_xs": qw_xs, "qu_xs": qu_xs, "px_zs": px_zs}

    def loss_function(self, x, fwd_rtn):
        r"""Wrapper function for mmVAE loss.

        Args:
            x (list): list of input data of type torch.Tensor.
            fwd_rtn (dict): dictionary containing encoding and decoding distributions.

        Returns:
            losses (dict): dictionary containing mmVAE loss.
        """
        qw_xs, qu_xs, px_zs = fwd_rtn["qw_xs"], fwd_rtn["qu_xs"], fwd_rtn["px_zs"]
        total = -self.moe_iwae(x, qw_xs, qu_xs, px_zs)
        losses = {"loss": total}
        return losses

    def moe_iwae(self, x, qw_xs, qu_xs, px_zs):
        r"""Calculate Mixture-of-Experts importance weighted autoencoder (IWAE) loss used for the mmVAE model.
        
        Args:
            x (list): list of input data of type torch.Tensor.
            fwd_rtn (dict): dictionary containing encoding and decoding distributions.

        Returns:
            (torch.Tensor): the output tensor.
        """
        lws = []
        prior = hydra.utils.instantiate(self.cfg.prior, loc=self.mean_priors_shared, logvar=self.logvar_priors_shared)
 
        for i in range(self.n_views):
            if self._training:
                us = qu_xs[i].rsample(torch.Size([self.K]))
                ws = qw_xs[i].rsample(torch.Size([self.K]))
                zs = torch.cat((us, ws), dim=-1)
            else:
                us = qu_xs[i]._sample()
                ws = qw_xs[i]._sample()
                zs = torch.cat((us, ws), dim=-1)

            lpz = prior.log_likelihood(zs).sum(-1)
            lqu_x = self.log_mean_exp(
                torch.stack([qu_x.log_likelihood(us).sum(-1) for qu_x in qu_xs])
            )  # summing over M modalities for each z to create q(u|x1:M)
            
            lqw_x = qw_xs[i].log_likelihood(ws).sum(-1)

            lpx_z = [
                px_z.log_likelihood(x[d]).view(*px_z._sample().size()[:2], -1).sum(-1)
                for d, px_z in enumerate(px_zs[i])
            ]  # summing over each decoder
            lpx_z = torch.stack(lpx_z).sum(0)

            lw = lpx_z + self.beta*(lpz - lqu_x - lqw_x)
            lws.append(lw)

        return (
            self.log_mean_exp(torch.stack(lws), dim=1).mean(0).sum()/self.n_views
          )  # looser iwae bound 
            

    def log_mean_exp(self, value, dim=0, keepdim=False):
        r"""Returns the log of the mean of the exponentials along the given dimension (dim).

        Args:
            value (torch.Tensor): the input tensor.
            dim (int, optional): the dimension along which to take the mean.
            keepdim (bool, optional): whether the output tensor has dim retained or not.

        Returns:
            (torch.Tensor): the output tensor.
        """
        return torch.logsumexp(value, dim, keepdim=keepdim) - math.log(value.size(dim))



import torch
import hydra

from ..base.constants import MODEL_MCVAE
from ..base.base_model import BaseModelVAE

class mcVAE(BaseModelVAE):
    r"""
    Multi-Channel Variational Autoencoder and Sparse Multi-Channel Variational Autoencoder.

    Code is based on: https://github.com/ggbioing/mcvae

    Args:
        cfg (str): Path to configuration file. Model specific parameters in addition to default parameters:
            
            - model.beta (int, float): KL divergence weighting term.
            - model.sparse (bool): Whether to enforce sparsity of the encoding distribution.
            - model.threshold (float): Dropout threshold applied to the latent dimensions. Default is 0.
            - encoder.default._target_ (multiviewae.architectures.mlp.VariationalEncoder): Type of encoder class to use.
            - encoder.default.enc_dist._target_ (multiviewae.base.distributions.Normal, multiviewae.base.distributions.MultivariateNormal): Encoding distribution.
            - decoder.default._target_ (multiviewae.architectures.mlp.VariationalDecoder): Type of decoder class to use.
            - decoder.default.init_logvar (int, float): Initial value for log variance of decoder.
            - decoder.default.dec_dist._target_ (multiviewae.base.distributions.Normal, multiviewae.base.distributions.MultivariateNormal): Decoding distribution.

        input_dim (list): Dimensionality of the input data.
        z_dim (int): Number of latent dimensions.

    References
    ----------
    Antelmi, Luigi & Ayache, Nicholas & Robert, Philippe & Lorenzi, Marco. (2019). Sparse Multi-Channel Variational Autoencoder for the Joint Analysis of Heterogeneous Data.
    """

    def __init__(
        self,
        cfg = None,
        input_dim = None,
        z_dim = None
    ):
        super().__init__(model_name=MODEL_MCVAE,
                        cfg=cfg,
                        input_dim=input_dim,
                        z_dim=z_dim)

    def encode(self, x):
        r"""Forward pass through encoder networks.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            qz_xs (list): list of encoding dimensions for each view.
        """
        qz_xs = []
        for i in range(self.n_views):
            mu, logvar = self.encoders[i](x[i])
            qz_x = hydra.utils.instantiate(
                eval(f"self.cfg.encoder.enc{i}.enc_dist"), loc=mu, logvar=logvar
            )
            qz_xs.append(qz_x)
        return qz_xs

    def decode(self, qz_xs):
        r"""Forward pass through decoder networks. Each latent is passed through all of the decoders.

        Args:
            z (list): list of latent dimensions for each view of type torch.Tensor.

        Returns:
            px_zs (list): A nested list of decoding distributions. The outer list has a n_view element indicating latent dimensions index. 
            The inner list is a n_view element list with the position in the list indicating the decoder index.
        """
        px_zs = []
        for qz_x in qz_xs:
            px_z = [
                self.decoders[i](qz_x._sample(training=self._training, return_mean=self.return_mean))
                for i in range(self.n_views)
            ]
            px_zs.append(px_z)
            del px_z
        return px_zs

    def forward(self, x):
        r"""Apply encode and decode methods to input data to generate latent dimensions and data reconstructions. 
        
        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            fwd_rtn (dict): dictionary containing encoding (qz_xs) and decoding (px_zs) distributions.
        """
        qz_xs = self.encode(x)
        px_zs = self.decode(qz_xs)
        fwd_rtn = {"px_zs": px_zs, "qz_xs": qz_xs}
        return fwd_rtn

    def loss_function(self, x, fwd_rtn):
        r"""Calculate mcVAE loss.

        Args:
            x (list): list of input data of type torch.Tensor.
            fwd_rtn (dict): dictionary containing encoding and decoding distributions.

        Returns:
            losses (dict): dictionary containing each element of the mcVAE loss.
        """
        px_zs = fwd_rtn["px_zs"]
        qz_xs = fwd_rtn["qz_xs"]
        kl = self.calc_kl(qz_xs)
        ll = self.calc_ll(x, px_zs)
        total = kl - ll
        losses = {"loss": total, "kl": kl, "ll": ll}
        return losses

    def calc_kl(self, qz_xs):
        r"""Calculate mcVAE KL-divergence loss.

        Args:
            qz_xs (list): list of encoding distributions.
            
        Returns:
            (torch.Tensor): KL-divergence loss across all views.
        """
        kl = 0
        for qz_x in qz_xs:
            if self.sparse:
                kl += qz_x.sparse_kl_divergence().mean(0).sum()
            else:
                kl += qz_x.kl_divergence(self.prior).mean(0).sum()
        return self.beta * kl

    def calc_ll(self, x, px_zs):
        r"""Calculate log-likelihood loss.

        Args:
            x (list): list of input data of type torch.Tensor.
            px_zs (list): list of decoding distributions.

        Returns:
            ll (torch.Tensor): Log-likelihood loss.
        """
        ll = 0
        for i in range(self.n_views):
            for j in range(self.n_views):
                ll += px_zs[j][i].log_likelihood(x[i]).mean(0).sum() #first index is latent, second index is view
        return ll/self.n_views/self.n_views



from .mlp import *
from .cnn import *

__all__ = ["Encoder", "VariationalEncoder", "ConditionalVariationalEncoder", 
           "Decoder", "VariationalDecoder", "ConditionalVariationalDecoder", 
           "Discriminator"]
classes = __all__ 



import torch

from ..base.constants import MODEL_MAAE, EPS
from ..base.base_model import BaseModelAAE

class mAAE(BaseModelAAE):
    r"""Multi-view Adversarial Autoencoder model with a separate latent representation for each view.

    Args:
        cfg (str): Path to configuration file. Model specific parameters in addition to default parameters:
            
            - eps (float): Value added for numerical stability.
            - discriminator._target_ (multiviewae.architectures.mlp.Discriminator): Discriminator network class.
            - discriminator.hidden_layer_dim (list): Number of nodes per hidden layer.
            - discriminator.bias (bool): Whether to include a bias term in hidden layers.
            - discriminator.non_linear (bool): Whether to include a ReLU() function between layers.
            - discriminator.dropout_threshold (float): Dropout threshold of layers.
        
        input_dim (list): Dimensionality of the input data.
        z_dim (int): Number of latent dimensions.
    """
    def __init__(
        self,
        cfg = None,
        input_dim = None,
        z_dim = None
    ):

        super().__init__(model_name=MODEL_MAAE,
                cfg=cfg,
                input_dim=input_dim,
                z_dim=z_dim)

    def encode(self, x):
        r"""Forward pass through encoder networks.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            z (list): list of latent dimensions for each view of type torch.Tensor.
        """
        z = []
        for i in range(self.n_views):
            z_ = self.encoders[i](x[i])
            z.append(z_)
        return z

    def decode(self, z):
        r"""Forward pass through decoder networks. Each latent is passed through all of the decoders.

        Args:
            z (list): list of latent dimensions for each view of type torch.Tensor.

        Returns:
            px_zs (list): list of decoding distributions.
        """
        px_zs = []
        for i in range(self.n_views):
            px_z = [self.decoders[j](z[i]) for j in range(self.n_views)]
            px_zs.append(px_z)
        return px_zs

    def disc(self, z):
        r"""Forward pass of "real" samples from gaussian prior and "fake" samples from encoders through the discriminator network.

        Args:
            z (list): list of latent dimensions for each view of type torch.Tensor.

        Returns:
            d_real (torch.Tensor): Discriminator network output for "real" samples.
            d_fake (list): list of discriminator network output for "fake" samples.
        """
        sh = z[0].shape
        z_real = self.prior.sample(sample_shape=sh)
        d_real = self.discriminator(z_real)
        d_fake = []
        for i in range(self.n_views):
            d = self.discriminator(z[i])
            d_fake.append(d)
        return d_real, d_fake

    def forward_recon(self, x):
        r"""Apply encode and decode methods to input data to generate latent dimensions and data reconstructions.
        
        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            fwd_rtn (dict): dictionary containing decoding distributions (px_zs) and latent dimensions (z).
        """
        z = self.encode(x)
        px_zs = self.decode(z)
        fwd_rtn = {"px_zs": px_zs, "z": z}
        return fwd_rtn

    def forward_discrim(self, x):
        r"""Apply encode and disc methods to input data to generate discriminator prediction on the latent dimensions and train discriminator parameters.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            fwd_rtn (dict): dictionary containing discriminator output from "real" samples (d_real), discriminator output from "fake" samples (d_fake), and latent dimensions (z).
        """
        [encoder.eval() for encoder in self.encoders]
        z = self.encode(x)
        d_real, d_fake = self.disc(z)
        fwd_rtn = {"d_real": d_real, "d_fake": d_fake, "z": z}
        return fwd_rtn

    def forward_gen(self, x):
        r"""Apply encode and disc methods to input data to generate discriminator prediction on the latent dimensions and train encoder parameters.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            fwd_rtn (dict): fwd_rtn (dict): dictionary containing discriminator output from "fake" samples (d_fake) and latent dimensions (z).
        """
        [encoder.train() for encoder in self.encoders]
        self.discriminator.eval()
        z = self.encode(x)
        _, d_fake = self.disc(z)
        fwd_rtn = {"d_fake": d_fake, "z": z}
        return fwd_rtn

    def recon_loss(self, x, fwd_rtn):
        r"""Calculate reconstruction loss.

        Args:
            x (list): list of input data of type torch.Tensor.
            fwd_rtn (dict): fwd_rtn from the forward_recon method.

        Returns:
            ll (torch.Tensor): Reconstruction error.
        """
        px_zs = fwd_rtn["px_zs"]
        ll = 0
        for i in range(self.n_views):
            for j in range(self.n_views):
                ll += - px_zs[j][i].log_likelihood(x[i]).mean(0).sum() #first index is latent, second index is view
        return ll / self.n_views / self.n_views

    def generator_loss(self, fwd_rtn):
        r"""Calculate the generator loss.

        Args:
            fwd_rtn (dict): fwd_rtn from the forward_gen method.

        Returns:
            gen_loss (torch.Tensor): Generator loss.
        """
        d_fake = fwd_rtn["d_fake"]
        gen_loss = 0
        for i in range(self.n_views):
            gen_loss += torch.mean(1 - torch.log(d_fake[i] + EPS)) 
        return gen_loss/self.n_views

    def discriminator_loss(self, fwd_rtn):
        r"""Calculate the discriminator loss.

        Args:
            fwd_rtn (dict): fwd_rtn from the forward_discrim method.

        Returns:
            disc_loss (torch.Tensor): Discriminator loss.
        """
        d_real = fwd_rtn["d_real"]
        d_fake = fwd_rtn["d_fake"]

        disc_loss = -torch.mean(torch.log(d_real + EPS))
        for i in range(self.n_views):
            disc_loss += -torch.mean(1 - torch.log(d_fake[i] + EPS)) 
        return disc_loss / (self.n_views + 1)





import torch
import hydra


import torch.nn as nn
import torch.nn.functional as F


class Encoder(nn.Module):
    """
    Configurable convolutional encoder.

    Args:
        input_dim (list): Dimensionality of the input data.
        num_filters (list[int]): Number of filters for each convolutional layer.
        input_shape (list[int]): Input shape to first conv layer.
        kernel_size (list[int]): kernel_size
        stride (list[int])
        padding (list[int])
        padding_mode (str)

    """
    def __init__(
        self,
        input_dim,
        z_dim,
        non_linear,
        bias,
        enc_dist,
        **kwargs
    ):
        super().__init__()

        self.input_size = input_dim
        self.z_dim = z_dim

        self.bias = bias
        self.non_linear = non_linear
        self.enc_dist = enc_dist

        self.layer_sizes = []

        conv_params = {k:v for (k,v) in kwargs.items() if k.startswith("layer")}

        layers = []

        num_layers = len(conv_params)
        for k,v in conv_params.items():
            l = v["layer"]
            v.pop('layer', None)
            layers.append(eval(f"nn.{l}(**v)"))

            out_size = [v1 for k1,v1 in v.items() if 'out_' in k1]
            if len(out_size) > 0:
                self.layer_sizes.append(out_size[0])

        # z_dim layer
        layers.append(nn.Linear(in_features=self.layer_sizes[-1], out_features=z_dim, bias=self.bias))
        self.layer_sizes.append(z_dim)

        self.encoder_layers = nn.Sequential(*layers)

    def forward(self, x):
        h1 = x
        for it_layer, layer in enumerate(self.encoder_layers[0:-1]):
            h1 = layer(h1)
            if self.non_linear:
                h1 = F.relu(h1)
        h1 = self.encoder_layers[-1](h1)
        return h1

class VariationalEncoder(Encoder):
    def __init__(
        self,
        input_dim,
        z_dim,
        non_linear,
        bias,
        sparse,
        log_alpha,
        enc_dist,
        **kwargs
    ):
        super().__init__(input_dim=input_dim,
                        z_dim=z_dim,
                        bias=bias,
                        non_linear=non_linear,
                        enc_dist=enc_dist,
                        **kwargs)

        self.sparse = sparse
        self.non_linear = non_linear
        self.log_alpha = log_alpha

        self.encoder_layers = self.encoder_layers[:-1]
        self.enc_mean_layer = nn.Linear(
            self.layer_sizes[-2],
            self.layer_sizes[-1],
            bias=self.bias,
        )

        if not self.sparse:
            self.enc_logvar_layer = nn.Linear(
                self.layer_sizes[-2],
                self.layer_sizes[-1],
                bias=self.bias,
            )

    def forward(self, x):
        h1 = x
        for it_layer, layer in enumerate(self.encoder_layers):
            h1 = layer(h1)
            if self.non_linear:
                h1 = F.relu(h1)

        if not self.sparse:
            mu = self.enc_mean_layer(h1)
            logvar = self.enc_logvar_layer(h1)
        else:
            mu = self.enc_mean_layer(h1)
            logvar = self.log_alpha + 2 * torch.log(torch.abs(mu) + 1e-8)

        return mu, logvar

class Decoder(nn.Module):
    def __init__(
        self,
        input_dim,
        z_dim,
        non_linear,
        bias,
        dec_dist,
        init_logvar=None,
        **kwargs
    ):
        super().__init__()

        self.input_size = input_dim
        self.z_dim = z_dim

        self.bias = bias
        self.dec_dist = dec_dist
        self.non_linear = non_linear

        self.layer_sizes = []
        self.layer_sizes.append(z_dim)

        conv_params = {k:v for (k,v) in kwargs.items() if k.startswith("layer")}

        layers = []

        # z_dim layer
        first_layer = list(conv_params)[0]
        out_dim = conv_params[first_layer]["out_features"]
        layers.append(nn.Linear(in_features=self.z_dim, out_features=out_dim, bias=self.bias))
        self.layer_sizes.append(out_dim)

        conv_params.pop(first_layer, None)
        num_layers = len(conv_params)
        for k,v in conv_params.items():
            l = v["layer"]
            v.pop('layer', None)
            layers.append(eval(f"nn.{l}(**v)")) 

            out_size = [v1 for k1,v1 in v.items() if 'out_' in k1]
            if len(out_size) > 0:
                self.layer_sizes.append(out_size[0])

        self.decoder_layers = nn.Sequential(*layers)

    def forward(self, z):
        x_rec = z
        for it_layer, layer in enumerate(self.decoder_layers[:-1]):
            x_rec = layer(x_rec)
            if self.non_linear:
                x_rec = F.relu(x_rec)

        x_rec = self.decoder_layers[-1](x_rec)
        x_rec = hydra.utils.instantiate(self.dec_dist, x=x_rec)

        return x_rec



import torch
import hydra
from ..base.constants import MODEL_DVCCA
from ..base.base_model import BaseModelVAE

class DVCCA(BaseModelVAE):
    r"""Deep Variational Canonical Correlation Analysis (DVCCA).

    Args:
        cfg (str): Path to configuration file. Model specific parameters in addition to default parameters:
            
            - model.beta (int, float): KL divergence weighting term.
            - model.private (bool): Whether to include private view-specific latent dimensions.
            - model.sparse (bool): Whether to enforce sparsity of the encoding distribution.
            - model.threshold (float): Dropout threshold applied to the latent dimensions. Default is 0.
            - encoder.default._target_ (multiviewae.architectures.mlp.VariationalEncoder): Type of encoder class to use.
            - encoder.default.enc_dist._target_ (multiviewae.base.distributions.Normal, multiviewae.base.distributions.MultivariateNormal): Encoding distribution.
            - decoder.default._target_ (multiviewae.architectures.mlp.VariationalDecoder): Type of decoder class to use.
            - decoder.default.init_logvar(int, float): Initial value for log variance of decoder.
            - decoder.default.dec_dist._target_ (multiviewae.base.distributions.Normal, multiviewae.base.distributions.MultivariateNormal): Decoding distribution.
        
        input_dim (list): Dimensionality of the input data.
        z_dim (int): Number of latent dimensions.

    References
    ----------
    Wang, Weiran & Lee, Honglak & Livescu, Karen. (2016). Deep Variational Canonical Correlation Analysis.


    """
    def __init__(
        self,
        cfg = None,
        input_dim = None,
        z_dim = None
    ):

        super().__init__(model_name=MODEL_DVCCA,
                        cfg=cfg,
                        input_dim=input_dim,
                        z_dim=z_dim)

    ################################            protected methods
    def _setencoders(self):
        r"""Set the encoder network using the first data input. If private=True also set a private encoder network for each view.
        """
        if self.sparse and self.threshold != 0.:
            self.log_alpha = torch.nn.Parameter(
                torch.FloatTensor(1, self.z_dim).normal_(0, 0.01)
            )
        else:
            self.sparse = False
            self.log_alpha = None

        self.encoders = torch.nn.ModuleList(
            [
                hydra.utils.instantiate(
                    self.cfg.encoder.default,  
                    input_dim=self.input_dim[0],
                    z_dim=self.z_dim,
                    sparse=self.sparse,
                    log_alpha=self.log_alpha,
                    _recursive_=False,
                    _convert_="all"
                )
            ]
        )

        if self.private:

            self.private_encoders = torch.nn.ModuleList(
                [
                    hydra.utils.instantiate(
                        eval(f"self.cfg.encoder.enc{i}"),
                        input_dim=d,
                        z_dim=self.z_dim,
                        sparse=self.sparse,
                        log_alpha=self.log_alpha,
                        _recursive_=False,
                        _convert_="all"
                    )
                    for i, d in enumerate(self.input_dim)
                ]
            )
            self.z_dim = self.z_dim + self.z_dim
            if self.sparse and self.threshold != 0.:

                self.log_alpha = torch.nn.Parameter(
                    torch.FloatTensor(1, self.z_dim).normal_(0, 0.01)
                )

    def configure_optimizers(self):
        r"""Configure optimizers for encoder, private encoder, and decoder network parameters.

        Returns:
            optimizers (list): list of Adam optimizers for encoders and decoders.
        """
        if self.private:
            optimizers = [
                torch.optim.Adam(self.encoders[0].parameters(), lr=self.learning_rate)
            ] + [
                torch.optim.Adam(
                    list(self.private_encoders[i].parameters()), lr=self.learning_rate
                )
                for i in range(self.n_views)
            ] + [
                torch.optim.Adam(
                    list(self.decoders[i].parameters()), lr=self.learning_rate
                )
                for i in range(self.n_views)
            ]
        else:
            optimizers = [
                torch.optim.Adam(self.encoders[0].parameters(), lr=self.learning_rate)
            ] + [
                torch.optim.Adam(
                    list(self.decoders[i].parameters()), lr=self.learning_rate
                )
                for i in range(self.n_views)
            ]
        return optimizers

    def encode(self, x):
        r"""Forward pass through encoder network. For DVCCA-private a forward pass is performed through each private encoder and the output latent is concatenated with the shared latent.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            Returns a combination of the following depending on the training stage and model type: 
            qz_x (list): list containing the shared encoding distribution.
            qz_xs (list): list of encoding distributions for shared and private latents of DVCCA-private.
            qh_xs (list): list of encoding distributions for private latents of DVCCA-private.

        """
        mu, logvar = self.encoders[0](x[0])
        qz_x = hydra.utils.instantiate(
                    eval(f"self.cfg.encoder.enc0.enc_dist"), loc=mu, logvar=logvar
        )
        if self.private:
            qz_xs = []
            qh_xs = []
            for i in range(self.n_views):
                mu_p, logvar_p = self.private_encoders[i](x[i])
                qh_x = hydra.utils.instantiate(
                    eval(f"self.cfg.encoder.enc{i}.enc_dist"), loc=mu_p, logvar=logvar_p
                )
                qh_xs.append(qh_x)

                mu_ = torch.cat((mu, mu_p), 1)
                logvar_ = torch.cat((logvar, logvar_p), 1)

                qz_x = hydra.utils.instantiate(
                    eval(f"self.cfg.encoder.enc{i}.enc_dist"), loc=mu_, logvar=logvar_
                )
                qz_xs.append(qz_x)
            if self._training:
                return [[qz_x], qz_xs, qh_xs]
            return qz_xs
        else:
            qz_x = hydra.utils.instantiate( 
                self.cfg.encoder.default.enc_dist, loc=mu, logvar=logvar
            )
            return [qz_x]

    def decode(self, qz_x):
        r"""Forward pass through decoder networks.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            (list): A nested list of decoding distributions, px_zs. The outer list has a single element indicating the shared or shared and private latent dimensions. 
            The inner list is a n_view element list with the position in the list indicating the decoder index.
        """
        px_zs = []
        for i in range(self.n_views):
            if self.private:
                x_out = self.decoders[i](qz_x[i]._sample(training=self._training, return_mean=self.return_mean))
            else:
                x_out = self.decoders[i](qz_x[0]._sample(training=self._training, return_mean=self.return_mean))
            px_zs.append(x_out)
        return [px_zs] 

    def forward(self, x):
        r"""Apply encode and decode methods to input data to generate latent dimensions and data reconstructions. 
        For DVCCA, the shared encoding distribution is passed to the decode method. 
        For DVCCA-private, the joint distribution of the shared and private latents for each view is passed to the decode method. 

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            fwd_rtn (dict): dictionary containing list of decoding distributions (px_zs), shared encoding distribution (qz_x), and (for DVCCA-private) private encoding distributions (qh_xs).
        """
        self.zero_grad()
        if self.private:
            qz_x, qz_xs, qh_xs = self.encode(x)
            px_zs = self.decode(qz_xs)
        else:
            qz_x = self.encode(x)
            px_zs = self.decode(qz_x)
            qh_xs = []
        fwd_rtn = {"px_zs": px_zs, "qz_x": qz_x, 'qh_xs': qh_xs}
        return fwd_rtn

    def loss_function(self, x, fwd_rtn):
        r"""Calculate DVCCA loss.
        
        Args:
            x (list): list of input data of type torch.Tensor.
            fwd_rtn (dict): dictionary containing list of decoding distributions (px_zs), shared encoding distribution (qz_x), and (for DVCCA-private) private encoding distributions (qh_xs).

        Returns:
            losses (dict): dictionary containing each element of the DVCCA loss.
        """
        px_zs = fwd_rtn["px_zs"]
        qz_x = fwd_rtn["qz_x"]
        qh_xs= fwd_rtn["qh_xs"]
        kl = self.calc_kl(qz_x, qh_xs)
        ll = self.calc_ll(x, px_zs)
        total = kl - ll
        losses = {"loss": total, "kl": kl, "ll": ll}
        return losses

    def calc_kl(self, qz_x, qh_xs):
        r"""Wrapper function for calculating KL-divergence loss.

        Args:
            qz_x (list): Single element list containing shared encoding distribution.
            qh_xs (list): list of  private encoding distributions for DVCCA-private.

        Returns:
            (torch.Tensor): KL-divergence loss across all views.
        """
        kl = 0
        kl += self.calc_kl_(qz_x[0])
        if self.private:      
            for i in range(self.n_views):
                kl+= self.calc_kl_(qh_xs[i])
        return self.beta * kl

    def calc_kl_(self, dist):
        r"""Calculate KL-divergence.

        Args:
            dist: Distribution object.

        Returns:
            (torch.Tensor): Kl-divergence.
        """
        if self.sparse:
            return dist.sparse_kl_divergence().mean(0).sum()
        return dist.kl_divergence(self.prior).mean(0).sum()

    def calc_ll(self, x, px_zs):
        r"""Calculate log-likelihood loss.

        Args:
            x (list): list of input data of type torch.Tensor.
            px_zs (list): list of decoding distributions.

        Returns:
            ll (torch.Tensor): Log-likelihood loss.
        """
        ll = 0
        for i in range(self.n_views):
            ll += px_zs[0][i].log_likelihood(x[i]).mean(0).sum()  #first index is latent, second index is view
        return ll
